
################################################################################
#                            skforecast.datasets                               #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

import pandas as pd
import textwrap
from urllib.parse import urlparse


def fetch_dataset(
    name: str,
    version: str = 'latest',
    raw: bool = False,
    kwargs_read_csv: dict = {},
    verbose: bool = True
) -> pd.DataFrame:
    """
    Fetch a dataset from the skforecast-datasets repository.

    Parameters
    ----------
    name: str
        Name of the dataset to fetch.
    version: str, int, default `'latest'`
        Version of the dataset to fetch. If 'latest', the latest version will be 
        fetched (the one in the main branch). For a list of available versions, 
        see the repository branches.
    raw: bool, default `False`
        If True, the raw dataset is fetched. If False, the preprocessed dataset 
        is fetched. The preprocessing consists of setting the column with the 
        date/time as index and converting the index to datetime. A frequency is 
        also set to the index.
    kwargs_read_csv: dict, default `{}`
        Kwargs to pass to pandas `read_csv` function.
    verbose: bool, default `True`
        If True, print information about the dataset.
    
    Returns
    -------
    df: pandas DataFrame
        Dataset.
    
    """

    version = 'main' if version == 'latest' else f'{version}'

    datasets = {
        'h2o': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/h2o.csv'
            ),
            'sep': ',',
            'index_col': 'fecha',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'file_type': 'csv',
            'description': (
                'Monthly expenditure ($AUD) on corticosteroid drugs that the '
                'Australian health system had between 1991 and 2008. '
            ),
            'source': (
                'Hyndman R (2023). fpp3: Data for Forecasting: Principles and Practice'
                '(3rd Edition). http://pkg.robjhyndman.com/fpp3package/,'
                'https://github.com/robjhyndman/fpp3package, http://OTexts.com/fpp3.'
            )
        },
        'h2o_exog': {
            'url': (
                f"https://raw.githubusercontent.com/skforecast/"
                f"skforecast-datasets/{version}/data/h2o_exog.csv"
            ),
            'sep': ',',
            'index_col': 'fecha',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'file_type': 'csv',
            'description': (
                'Monthly expenditure ($AUD) on corticosteroid drugs that the '
                'Australian health system had between 1991 and 2008. Two additional '
                'variables (exog_1, exog_2) are simulated.'
            ),
            'source': (
                "Hyndman R (2023). fpp3: Data for Forecasting: Principles and Practice "
                "(3rd Edition). http://pkg.robjhyndman.com/fpp3package/, "
                "https://github.com/robjhyndman/fpp3package, http://OTexts.com/fpp3."
            )
        },
        'fuel_consumption': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/consumos-combustibles-mensual.csv'
            ),
            'sep': ',',
            'index_col': 'Fecha',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'file_type': 'csv',
            'description': (
                'Monthly fuel consumption in Spain from 1969-01-01 to 2022-08-01.'
            ),
            'source': (
                'Obtained from Corporación de Reservas Estratégicas de Productos '
                'Petrolíferos and Corporación de Derecho Público tutelada por el '
                'Ministerio para la Transición Ecológica y el Reto Demográfico. '
                'https://www.cores.es/es/estadisticas'
            )
        },
        'items_sales': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/simulated_items_sales.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'csv',
            'description': 'Simulated time series for the sales of 3 different items.',
            'source': 'Simulated data.'
        },
        'air_quality_valencia': {
            'url': (
                f"https://raw.githubusercontent.com/skforecast/"
                f"skforecast-datasets/{version}/data/air_quality_valencia.csv"
            ),
            'sep': ',',
            'index_col': 'datetime',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'csv',
            'description': (
                'Hourly measures of several air chemical pollutant at Valencia city '
                '(Avd. Francia) from 2019-01-01 to 20213-12-31. Including the following '
                'variables: pm2.5 (µg/m³), CO (mg/m³), NO (µg/m³), NO2 (µg/m³), '
                'PM10 (µg/m³), NOx (µg/m³), O3 (µg/m³), Veloc. (m/s), Direc. (degrees), '
                'SO2 (µg/m³).'
            ),
            'source': (
                "Red de Vigilancia y Control de la Contaminación Atmosférica, "
                "46250047-València - Av. França, "
                "https://mediambient.gva.es/es/web/calidad-ambiental/datos-historicos."
            )
        },
        'air_quality_valencia_no_missing': {
            'url': (
                f"https://raw.githubusercontent.com/skforecast/"
                f"skforecast-datasets/{version}/data/air_quality_valencia_no_missing.csv"
            ),
            'sep': ',',
            'index_col': 'datetime',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'csv',
            'description': (
                'Hourly measures of several air chemical pollutant at Valencia city '
                '(Avd. Francia) from 2019-01-01 to 20213-12-31. Including the following '
                'variables: pm2.5 (µg/m³), CO (mg/m³), NO (µg/m³), NO2 (µg/m³), '
                'PM10 (µg/m³), NOx (µg/m³), O3 (µg/m³), Veloc. (m/s), Direc. (degrees), '
                'SO2 (µg/m³). Missing values have been imputed using linear interpolation.'
            ),
            'source': (
                "Red de Vigilancia y Control de la Contaminación Atmosférica, "
                "46250047-València - Av. França, "
                "https://mediambient.gva.es/es/web/calidad-ambiental/datos-historicos."
            )
        },
        'website_visits': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/visitas_por_dia_web_cienciadedatos.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': '1D',
            'file_type': 'csv',
            'description': (
                'Daily visits to the cienciadedatos.net website registered with the '
                'google analytics service.'
            ),
            'source': (
                "Amat Rodrigo, J. (2021). cienciadedatos.net (1.0.0). Zenodo. "
                "https://doi.org/10.5281/zenodo.10006330"
            )
        },
        'bike_sharing': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/bike_sharing_dataset_clean.csv'
            ),
            'sep': ',',
            'index_col': 'date_time',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'csv',
            'description': (
                'Hourly usage of the bike share system in the city of Washington D.C. '
                'during the years 2011 and 2012. In addition to the number of users per '
                'hour, information about weather conditions and holidays is available.'
            ),
            'source': (
                "Fanaee-T,Hadi. (2013). Bike Sharing Dataset. UCI Machine Learning "
                "Repository. https://doi.org/10.24432/C5W894."
            )
        },
        'bike_sharing_extended_features': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/bike_sharing_extended_features.csv'
            ),
            'sep': ',',
            'index_col': 'date_time',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'csv',
            'description': (
                'Hourly usage of the bike share system in the city of Washington D.C. '
                'during the years 2011 and 2012. In addition to the number of users per '
                'hour, the dataset was enriched by introducing supplementary features. '
                'Addition includes calendar-based variables (day of the week, hour of '
                'the day, month, etc.), indicators for sunlight, incorporation of '
                'rolling temperature averages, and the creation of polynomial features '
                'generated from variable pairs. All cyclic variables are encoded using '
                'sine and cosine functions to ensure accurate representation.'
            ),
            'source': (
                "Fanaee-T,Hadi. (2013). Bike Sharing Dataset. UCI Machine Learning "
                "Repository. https://doi.org/10.24432/C5W894."
            )
        },
        'australia_tourism': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/australia_tourism.csv'
            ),
            'sep': ',',
            'index_col': 'date_time',
            'date_format': '%Y-%m-%d',
            'freq': 'Q',
            'file_type': 'csv',
            'description': (
                "Quarterly overnight trips (in thousands) from 1998 Q1 to 2016 Q4 "
                "across Australia. The tourism regions are formed through the "
                "aggregation of Statistical Local Areas (SLAs) which are defined by "
                "the various State and Territory tourism authorities according to "
                "their research and marketing needs."
            ),
            'source': (
                "Wang, E, D Cook, and RJ Hyndman (2020). A new tidy data structure to "
                "support exploration and modeling of temporal data, Journal of "
                "Computational and Graphical Statistics, 29:3, 466-478, "
                "doi:10.1080/10618600.2019.1695624."
            )
        },
        'uk_daily_flights': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/uk_daily_flights.csv'
            ),
            'sep': ',',
            'index_col': 'Date',
            'date_format': '%d/%m/%Y',
            'freq': 'D',
            'file_type': 'csv',
            'description': 'Daily number of flights in UK from 02/01/2019 to 23/01/2022.',
            'source': (
                'Experimental statistics published as part of the Economic activity and '
                'social change in the UK, real-time indicators release, Published 27 '
                'January 2022. Daily flight numbers are available in the dashboard '
                'provided by the European Organisation for the Safety of Air Navigation '
                '(EUROCONTROL). '
                'https://www.ons.gov.uk/economy/economicoutputandproductivity/output/'
                'bulletins/economicactivityandsocialchangeintheukrealtimeindicators/latest'
            )
        },
        'wikipedia_visits': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/wikipedia_visits.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'csv',
            'description': (
                'Log daily page views for the Wikipedia page for Peyton Manning. '
                'Scraped data using the Wikipediatrend package in R.'
            ),
            'source': (
                'https://github.com/facebook/prophet/blob/main/examples/'
                'example_wp_log_peyton_manning.csv'
            )
        },
        'vic_electricity': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/vic_electricity.csv'
            ),
            'sep': ',',
            'index_col': 'Time',
            'date_format': '%Y-%m-%dT%H:%M:%SZ',
            'freq': '30min',
            'file_type': 'csv',
            'description': 'Half-hourly electricity demand for Victoria, Australia',
            'source': (
                "O'Hara-Wild M, Hyndman R, Wang E, Godahewa R (2022).tsibbledata: Diverse "
                "Datasets for 'tsibble'. https://tsibbledata.tidyverts.org/, "
                "https://github.com/tidyverts/tsibbledata/. "
                "https://tsibbledata.tidyverts.org/reference/vic_elec.html"
            )
        },
        'store_sales': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/store_sales.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'csv',
            'description': (
                'This dataset contains 913,000 sales transactions from 2013-01-01 to '
                '2017-12-31 for 50 products (SKU) in 10 stores.'
            ),
            'source': (
                'The original data was obtained from: inversion. (2018). Store Item '
                'Demand Forecasting Challenge. Kaggle. '
                'https://kaggle.com/competitions/demand-forecasting-kernels-only'
            )
        },
        'bicimad': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/bicimad_users.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'csv',
            'description': (
                'This dataset contains the daily users of the bicycle rental '
                'service (BiciMad) in the city of Madrid (Spain) from 2014-06-23 '
                'to 2022-09-30.'
            ),
            'source': (
                'The original data was obtained from: Portal de datos abiertos '
                'del Ayuntamiento de Madrid https://datos.madrid.es/portal/site/egob'
            )
        },
        'm4_daily': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/m4_daily.parquet'
            ),
            'sep': None,
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'D',
            'file_type': 'parquet',
            'description': "Time series with daily frequency from the M4 competition.",
            'source': (
                "Monash Time Series Forecasting Repository  "
                "(https://zenodo.org/communities/forecasting) Godahewa, R., Bergmeir, "
                "C., Webb, G. I., Hyndman, R. J., & Montero-Manso, P. (2021). Monash "
                "Time Series Forecasting Archive. In Neural Information Processing "
                "Systems Track on Datasets and Benchmarks. \n"
                "Raw data, available in .tsf format, has been converted to Pandas "
                "format using the code provided by the authors in "
                "https://github.com/rakshitha123/TSForecasting/blob/master/utils/data_loader.py \n"
                "The category of each time series has been included in the dataset. This "
                "information has been obtained from the Kaggle competition page: "
                "https://www.kaggle.com/datasets/yogesh94/m4-forecasting-competition-dataset"
            )
        },
        'm4_hourly': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/m4_hourly.parquet'
            ),
            'sep': None,
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'parquet',
            'description': "Time series with hourly frequency from the M4 competition.",
            'source': (
                "Monash Time Series Forecasting Repository  "
                "(https://zenodo.org/communities/forecasting) Godahewa, R., Bergmeir, "
                "C., Webb, G. I., Hyndman, R. J., & Montero-Manso, P. (2021). Monash "
                "Time Series Forecasting Archive. In Neural Information Processing "
                "Systems Track on Datasets and Benchmarks. \n"
                "Raw data, available in .tsf format, has been converted to Pandas "
                "format using the code provided by the authors in "
                "https://github.com/rakshitha123/TSForecasting/blob/master/utils/data_loader.py \n"
                "The category of each time series has been included in the dataset. This "
                "information has been obtained from the Kaggle competition page: "
                "https://www.kaggle.com/datasets/yogesh94/m4-forecasting-competition-dataset"
            )
        },
        'ashrae_daily': {
            'url': 'https://drive.google.com/file/d/1fMsYjfhrFLmeFjKG3jenXjDa5s984ThC/view?usp=sharing',
            'sep': None,
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'parquet',
            'description': (
                "Daily energy consumption data from the ASHRAE competition with "
                "building metadata and weather data."
            ),
            'source': (
                "Kaggle competition Addison Howard, Chris Balbach, Clayton Miller, "
                "Jeff Haberl, Krishnan Gowri, Sohier Dane. (2019). ASHRAE - Great Energy "
                "Predictor III. Kaggle. https://www.kaggle.com/c/ashrae-energy-prediction/overview"
            )
        },
        'bdg2_daily': {
            'url': 'https://drive.google.com/file/d/1KHYopzclKvS1F6Gt6GoJWKnxiuZ2aqen/view?usp=sharing',
            'sep': None,
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'parquet',
            'description': (
                "Daily energy consumption data from the The Building Data Genome Project 2 "
                "with building metadata and weather data. "
                "https://github.com/buds-lab/building-data-genome-project-2"
            ),
            'source': (
                "Miller, C., Kathirgamanathan, A., Picchetti, B. et al. The Building Data "
                "Genome Project 2, energy meter data from the ASHRAE Great Energy "
                "Predictor III competition. Sci Data 7, 368 (2020). "
                "https://doi.org/10.1038/s41597-020-00712-x"
            )
        },
        'bdg2_daily_sample': {
            'url': 'https://raw.githubusercontent.com/skforecast/skforecast-datasets/refs/heads/main/data/bdg2_daily_sample.csv',
            'sep': ',',
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'csv',
            'description': (
                "Daily energy consumption data of two buildings sampled from the "
                "The Building Data Genome Project 2. "
                "https://github.com/buds-lab/building-data-genome-project-2"
            ),
            'source': (
                "Miller, C., Kathirgamanathan, A., Picchetti, B. et al. The Building Data "
                "Genome Project 2, energy meter data from the ASHRAE Great Energy "
                "Predictor III competition. Sci Data 7, 368 (2020). "
                "https://doi.org/10.1038/s41597-020-00712-x"
            )
        },
        'bdg2_hourly': {
            'url': 'https://drive.google.com/file/d/1I2i5mZJ82Cl_SHPTaWJmLoaXnntdCgh7/view?usp=sharing',
            'sep': None,
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'parquet',
            'description': (
                "Hourly energy consumption data from the The Building Data Genome Project 2 "
                "with building metadata and weather data. "
                "https://github.com/buds-lab/building-data-genome-project-2"
            ),
            'source': (
                "Miller, C., Kathirgamanathan, A., Picchetti, B. et al. The Building Data "
                "Genome Project 2, energy meter data from the ASHRAE Great Energy "
                "Predictor III competition. Sci Data 7, 368 (2020). "
                "https://doi.org/10.1038/s41597-020-00712-x"
            )
        },
        'bdg2_hourly_sample': {
            'url': 'https://raw.githubusercontent.com/skforecast/skforecast-datasets/refs/heads/main/data/bdg2_hourly_sample.csv',
            'sep': ',',
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'H',
            'file_type': 'csv',
            'description': (
                "Daily energy consumption data of two buildings sampled from the "
                "The Building Data Genome Project 2. "
                "https://github.com/buds-lab/building-data-genome-project-2"
            ),
            'source': (
                "Miller, C., Kathirgamanathan, A., Picchetti, B. et al. The Building Data "
                "Genome Project 2, energy meter data from the ASHRAE Great Energy "
                "Predictor III competition. Sci Data 7, 368 (2020). "
                "https://doi.org/10.1038/s41597-020-00712-x"
            )
        },
        'm5': {
            'url': [
                'https://drive.google.com/file/d/1JOqBsSHegly6iSJFgmkugAko734c6ZW5/view?usp=sharing',
                'https://drive.google.com/file/d/1BhO1BUvs-d7ipXrm7caC3Wd_d0C_6PZ8/view?usp=sharing',
                'https://drive.google.com/file/d/1oHwkQ_QycJVTZMb6bH8C2klQB971gXXA/view?usp=sharing',
                'https://drive.google.com/file/d/1OvYzFlDG04YgTvju2k02vHEOj0nIuwei/view?usp=sharing'
            ],
            'sep': None,
            'index_col': 'timestamp',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'parquet',
            'description': (
                "Daily sales data from the M5 competition with product metadata and calendar data."
            ),
            'source': (
                "Addison Howard, inversion, Spyros Makridakis, and vangelis. "
                "M5 Forecasting - Accuracy. https://kaggle.com/competitions/m5-forecasting-accuracy, 2020. Kaggle."
            )
        },
        'ett_m1': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/refs/heads/{version}/data/ETTm1.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': '15min',
            'file_type': 'csv',
            'description': (
                'Data from an electricity transformer station was collected between July '
                '2016 and July 2018 (2 years x 365 days x 24 hours x 4 intervals per '
                'hour = 70,080 data points). Each data point consists of 8 features, '
                'including the date of the point, the predictive value "Oil Temperature (OT)", '
                'and 6 different types of external power load features: High UseFul Load (HUFL), '
                'High UseLess Load (HULL), Middle UseFul Load (MUFL), Middle UseLess Load (MULL), '
                'Low UseFul Load (LUFL), Low UseLess Load (LULL).'
            ),
            'source': (
                'Zhou, Haoyi & Zhang, Shanghang & Peng, Jieqi & Zhang, Shuai & Li, '
                'Jianxin & Xiong, Hui & Zhang, Wancai. (2020). Informer: Beyond Efficient '
                'Transformer for Long Sequence Time-Series Forecasting. '
                '[10.48550/arXiv.2012.07436](https://arxiv.org/abs/2012.07436). '
                'https://github.com/zhouhaoyi/ETDataset'
            )
        },
        'ett_m2': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/ETTm2.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': '15min',
            'file_type': 'csv',
            'description': (
                'Data from an electricity transformer station was collected between July '
                '2016 and July 2018 (2 years x 365 days x 24 hours x 4 intervals per '
                'hour = 70,080 data points). Each data point consists of 8 features, '
                'including the date of the point, the predictive value "Oil Temperature (OT)", '
                'and 6 different types of external power load features: High UseFul Load (HUFL), '
                'High UseLess Load (HULL), Middle UseFul Load (MUFL), Middle UseLess Load (MULL), '
                'Low UseFul Load (LUFL), Low UseLess Load (LULL).'
            ),
            'source': (
                'Zhou, Haoyi & Zhang, Shanghang & Peng, Jieqi & Zhang, Shuai & Li, '
                'Jianxin & Xiong, Hui & Zhang, Wancai. (2020). Informer: Beyond Efficient '
                'Transformer for Long Sequence Time-Series Forecasting. '
                '[10.48550/arXiv.2012.07436](https://arxiv.org/abs/2012.07436). '
                'https://github.com/zhouhaoyi/ETDataset'
            )
        },
        'ett_m2_extended': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/'
                f'skforecast-datasets/{version}/data/ETTm2_extended.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': '15min',
            'file_type': 'csv',
            'description': (
                'Data from an electricity transformer station was collected between July '
                '2016 and July 2018 (2 years x 365 days x 24 hours x 4 intervals per '
                'hour = 70,080 data points). Each data point consists of 8 features, '
                'including the date of the point, the predictive value "Oil Temperature (OT)", '
                'and 6 different types of external power load features: High UseFul Load (HUFL), '
                'High UseLess Load (HULL), Middle UseFul Load (MUFL), Middle UseLess Load (MULL), '
                'Low UseFul Load (LUFL), Low UseLess Load (LULL). Additional variables are '
                'created based on calendar information (year, month, week, day of the week, '
                'and hour). These variables have been encoded using the cyclical encoding '
                'technique (sin and cos transformations) to preserve the cyclical nature '
                'of the data.'
            ),
            'source': (
                'Zhou, Haoyi & Zhang, Shanghang & Peng, Jieqi & Zhang, Shuai & Li, '
                'Jianxin & Xiong, Hui & Zhang, Wancai. (2020). Informer: Beyond Efficient '
                'Transformer for Long Sequence Time-Series Forecasting. '
                '[10.48550/arXiv.2012.07436](https://arxiv.org/abs/2012.07436). '
                'https://github.com/zhouhaoyi/ETDataset'
            )
        },
        'expenditures_australia': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/skforecast-datasets/refs/heads/'
                f'{version}/data/expenditures_australia.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'MS',
            'file_type': 'csv',
            'description': (
                'Monthly expenditure on cafes, restaurants and takeaway food services '
                'in Victoria (Australia) from April 1982 up to April 2024.'
            ),
            'source': (
                'Australian Bureau of Statistics. Catalogue No. 8501.0 '
                'https://www.abs.gov.au/statistics/industry/retail-and-wholesale-trade/'
                'retail-trade-australia/apr-2024/8501011.xlsx'
            )
        },
        'public_transport_madrid': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/skforecast-datasets/refs/heads/'
                f'{version}/data/public-transport-madrid.csv'
            ),
            'sep': ',',
            'index_col': 'date',
            'date_format': '%Y-%m-%d',
            'freq': 'D',
            'file_type': 'csv',
            'description': (
                'Daily users of public transport in Madrid (Spain) from 2023-01-01 to 2024-12-15.'
            ),
            'source': (
                'Consorcio Regional de Transportes de Madrid CRTM, CRTM Evolucion demanda diaria '
                'https://datos.crtm.es/documents/a7210254c4514a19a51b1617cfd61f75/about'
            )
        },
        'turbine_emission': {
            'url': (
                f'https://raw.githubusercontent.com/skforecast/skforecast-datasets/refs/heads/'
                f'{version}/data/turbine_emission.csv'
            ),
            'sep': ',',
            'index_col': 'datetime',
            'date_format': '%Y-%m-%d %H:%M:%S',
            'freq': 'h',
            'file_type': 'csv',
            'description': (
                'The dataset contains 36733 instances of 11 sensor measures aggregated over one hour, '
                'from a gas turbine located in Turkey for the purpose of studying flue gas emissions, '
                'namely CO and NOx. Available variables include: Ambient temperature (AT), '
                'Ambient pressure (AP), Ambient humidity (AH), Air filter difference pressure (AFDP), '
                'Gas turbine exhaust pressure (GTEP), Turbine inlet temperature (TIT), '
                'Turbine after temperature (TAT), Compressor discharge pressure (CDP), Turbine energy yield (TEY), '
                'Carbon monoxide (CO), and Nitrogen oxides (NOx). Covered period from 2011-01-01 00:00:00 '
                'to 2015-03-11 12:00:00.'
            ),
            'source': (
                'https://archive.ics.uci.edu/dataset/551/gas+turbine+co+and+nox+emission+data+set'
            )
        },
    }
    
    if name not in datasets.keys():
        raise ValueError(
            f"Dataset '{name}' not found. Available datasets are: {list(datasets.keys())}"
        )
    
    url = datasets[name]['url']
    file_type = datasets[name]['file_type']

    if not isinstance(url, list):
        parsed = urlparse(url)
        if parsed.scheme == "https" and parsed.netloc == "drive.google.com":
            file_id = url.split('/')[-2]
            url = 'https://drive.google.com/uc?id=' + file_id
        if file_type == 'csv':
            try:
                sep = datasets[name]['sep']
                df = pd.read_csv(url, sep=sep, **kwargs_read_csv)
            except Exception as e:
                raise ValueError(
                    f"Error reading dataset '{name}' from {url}: {str(e)}."
                )
        if file_type == 'parquet':
            try:
                df = pd.read_parquet(url)
            except Exception as e:
                raise ValueError(
                    f"Error reading dataset '{name}' from {url}: {str(e)}."
                )
    else:
        try: 
            df = []
            for url_partition in url:
                path = 'https://drive.google.com/uc?export=download&id=' + url_partition.split('/')[-2]
                df.append(pd.read_parquet(path))
        except Exception as e:
            raise ValueError(
                f"Error reading dataset '{name}' from {url}: {str(e)}."
            )
        df = pd.concat(df, axis=0).reset_index(drop=True)

    if not raw:
        try:
            index_col = datasets[name]['index_col']
            freq = datasets[name]['freq']
            if freq == 'H' and pd.__version__ >= '2.2.0':
                freq = "h"
            date_format = datasets[name]['date_format']
            df = df.set_index(index_col)
            df.index = pd.to_datetime(df.index, format=date_format)
            df = df.asfreq(freq)
            df = df.sort_index()
        except:
            pass
    
    if verbose:
        print(name)
        print('-' * len(name))
        description = textwrap.fill(datasets[name]['description'], width=80)
        source = textwrap.fill(datasets[name]['source'], width=80)
        print(description)
        print(source)
        print(f"Shape of the dataset: {df.shape}")

    return df


def load_demo_dataset(version: str = 'latest') -> pd.Series:
    """
    Load demo data set with monthly expenditure ($AUD) on corticosteroid drugs that
    the Australian health system had between 1991 and 2008. Obtained from the book:
    Forecasting: Principles and Practice by Rob J Hyndman and George Athanasopoulos.
    Index is set to datetime with monthly frequency and sorted.

    Parameters
    ----------
    version: str, default `'latest'`
        Version of the dataset to fetch. If 'latest', the latest version will be
        fetched (the one in the main branch). For a list of available versions,
        see the repository branches.

    Returns
    -------
    df: pandas Series
        Dataset.
    
    """

    version = 'main' if version == 'latest' else f'{version}'

    url = (
        f'https://raw.githubusercontent.com/skforecast/skforecast-datasets/{version}/'
        'data/h2o.csv'
    )

    df = pd.read_csv(url, sep=',', header=0, names=['y', 'datetime'])
    df['datetime'] = pd.to_datetime(df['datetime'], format='%Y-%m-%d')
    df = df.set_index('datetime')
    df = df.asfreq('MS')
    df = df['y']
    df = df.sort_index()

    return df



from ._forecaster_equivalent_date import ForecasterEquivalentDate
from ._forecaster_recursive import ForecasterRecursive
from ._forecaster_recursive_multiseries import ForecasterRecursiveMultiSeries
from ._forecaster_sarimax import ForecasterSarimax



# Unit test set_out_sample_residuals ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import StandardScaler
from skforecast.exceptions import ResidualsUsageWarning
from skforecast.direct import ForecasterDirectMultiVariate

# Fixtures
series = pd.DataFrame({'l1': pd.Series(np.arange(15)), 
                       'l2': pd.Series(np.arange(15))})


def test_set_out_sample_residuals_NotFittedError_when_forecaster_not_fitted():
    """
    Test NotFittedError is raised when forecaster is not fitted.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `set_out_sample_residuals()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_is_not_dict():
    """
    Test TypeError is raised when y_true is not a dict.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = 'not_dict'
    y_pred = {'l1': np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"`y_true` must be a dictionary of numpy ndarrays or pandas Series. "
        f"Got {type(y_true)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_is_not_dict():
    """
    Test TypeError is raised when y_pred is not a dict.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}
    y_pred = 'not_dict'

    err_msg = re.escape(
        f"`y_pred` must be a dictionary of numpy ndarrays or pandas Series. "
        f"Got {type(y_pred)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_pred_and_y_true_keys_do_not_match():
    """
    Test TypeError is raised when y_pred and y_true keys do not match.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 2: np.array([1, 2, 3, 4, 5])}
    y_pred = {3: np.array([1, 2, 3, 4, 5]), 4: np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"`y_true` and `y_pred` must have the same keys. "
        f"Got {set(y_true.keys())} and {set(y_pred.keys())}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_contains_no_numpy_ndarrays_or_pandas_series():
    """
    Test TypeError is raised when y_true contains no numpy ndarrays or pandas series.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': 'not_ndarray'}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"Values of `y_true` must be numpy ndarrays or pandas Series. "
        f"Got {type(y_true['l1'])} for series l1."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_contains_no_numpy_ndarrays_or_pandas_series():
    """
    Test TypeError is raised when y_pred contains no numpy ndarrays or pandas series.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': 'not_ndarray'}

    err_msg = re.escape(
        f"Values of `y_pred` must be numpy ndarrays or pandas Series. "
        f"Got {type(y_pred['l1'])} for series l1."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_elements_with_different_lengths():
    """
    Test ValueError is raised when y_true and y_pred have elements with different lengths.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2])}

    err_msg = re.escape(
        '`y_true` and `y_pred` must have the same length. Got 5 and 2 for series l2.'
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_series_with_different_indexes():
    """
    Test ValueError is raised when y_true and y_pred have series with different indexes.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': pd.Series([1, 2, 3, 4, 5], index=[1, 2, 3, 4, 5])}
    y_pred = {'l1': pd.Series([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "When containing pandas Series, elements in `y_true` and "
        "`y_pred` must have the same index."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_inputs_does_not_match_the_target_level():
    """
    Test ValueError is raised when inputs does not contain keys that match any step.
    """
    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.fit(series=series)
    y_true = {'l3': np.array([1, 2, 3])}
    y_pred = {'l3': np.array([1, 2, 3])}

    err_msg = re.escape(
        "`y_pred` and `y_true` must have only the key 'l1'. Got {'l3'}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_no_append():
    """
    Test residuals stored when new residuals length is less than 10_000 and 
    append is False.
    """
    rng = np.random.default_rng(123)
    y_true = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000)), 
    }
    y_pred = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000)), 
    }

    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3, transformer_series=None
    )
    forecaster.fit(series=series)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=False)
    results = {
        'l1': np.sort(forecaster.out_sample_residuals_['l1']),
    }

    expected = {
        'l1': np.sort(y_true['l1'] - y_pred['l1']),
    }

    assert forecaster.out_sample_residuals_.keys() == expected.keys()
    for key in results.keys():
        np.testing.assert_array_almost_equal(expected[key], results[key])


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_append():
    """
    Test residuals stored when new residuals length is less than 10_000 and 
    append is True.
    """
    rng = np.random.default_rng(123)
    y_true = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000))
    }
    y_pred = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000))
    }

    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=2, lags=3, transformer_series=None
    )
    forecaster.fit(series=series)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)
    results = {
        'l1': np.sort(forecaster.out_sample_residuals_['l1'])
    }

    residuals_1 = (y_true['l1'] - y_pred['l1'])
    expected = {
        'l1': np.sort(np.concatenate((residuals_1, residuals_1))),
    }

    assert forecaster.out_sample_residuals_.keys() == expected.keys()
    for key in results.keys():
        np.testing.assert_array_almost_equal(expected[key], results[key])


def test_set_out_sample_residuals_when_residuals_length_is_greater_than_10000():
    """
    Test length residuals stored when its length is greater than 10_000.
    """
    rng = np.random.RandomState(42)
    series_fit = pd.DataFrame(
        data = {
            'l1': rng.normal(loc=10, scale=1, size=50_000),
            'l2': rng.normal(loc=10, scale=1, size=50_000)
        }
    )

    forecaster = ForecasterDirectMultiVariate(
        regressor=LinearRegression(), level='l1', steps=1, lags=1, 
        transformer_series=None, binner_kwargs={"n_bins": 10}
    )
    forecaster.fit(series=series_fit)
    X_train, y_train = forecaster.create_train_X_y(series=series_fit)
    X_train_step_1, y_train_step_1 = forecaster.filter_train_X_y_for_step(
        step=1, X_train=X_train, y_train=y_train
    )

    y_true = {'l1': y_train_step_1}
    y_pred = {'l1': forecaster.regressors_[1].predict(X_train_step_1)}
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)

    assert list(forecaster.out_sample_residuals_.keys()) == ['l1']
    assert len(forecaster.out_sample_residuals_['l1']) == 10_000
    for v in forecaster.out_sample_residuals_by_bin_['l1'].values():
        assert len(v) == 1_000


def test_out_sample_residuals_by_bin_and_in_sample_residuals_by_bin_equivalence():
    """
    Test out sample residuals by bin are equivalent to in-sample residuals by bin
    when training data and training predictions are passed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor          = LinearRegression(),
                     level              = 'l1',
                     steps              = 3,
                     lags               = 5,
                     transformer_series = None,
                     binner_kwargs      = {'n_bins': 3}
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    X_train, y_train = forecaster.create_train_X_y(series=series)

    y_true = []
    y_pred = []
    for step in forecaster.steps:
        X_train_step, y_train_step = forecaster.filter_train_X_y_for_step(
            step=step, X_train=X_train, y_train=y_train
        )
        y_true.append(y_train_step.to_numpy())
        y_pred.append(forecaster.regressors_[step].predict(X_train_step))

    y_true = {'l1': np.concatenate(y_true)}
    y_pred = {'l1': np.concatenate(y_pred)}

    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)

    assert forecaster.in_sample_residuals_by_bin_.keys() == forecaster.out_sample_residuals_by_bin_.keys()
    assert forecaster.in_sample_residuals_by_bin_['l1'].keys() == forecaster.out_sample_residuals_by_bin_['l1'].keys()
    for k in forecaster.out_sample_residuals_by_bin_['l1'].keys():
        np.testing.assert_array_almost_equal(
            forecaster.in_sample_residuals_by_bin_['l1'][k],
            forecaster.out_sample_residuals_by_bin_['l1'][k]
        )


def test_set_out_sample_residuals_append_new_residuals_per_bin():
    """
    Test that set_out_sample_residuals append residuals per bin until it
    reaches the max allowed size of 10_000 // n_bins
    """
    rng = np.random.default_rng(12345)
    series_fit = pd.DataFrame(
        data = {
            'l1': rng.normal(loc=10, scale=1, size=1001),
            'l2': rng.normal(loc=10, scale=1, size=1001)
        },
        index = pd.date_range(start="01-01-2000", periods=1001, freq="h"),
    )

    forecaster = ForecasterDirectMultiVariate(
                     regressor          = LinearRegression(),
                     level              = 'l1',
                     steps              = 1,
                     lags               = 1,
                     transformer_series = None,
                     binner_kwargs      = {'n_bins': 2}
                 )
    forecaster.fit(series=series_fit)
    X_train, y_train = forecaster.create_train_X_y(series=series_fit)
    X_train_step_1, y_train_step_1 = forecaster.filter_train_X_y_for_step(
        step=1, X_train=X_train, y_train=y_train
    )

    y_true = {'l1': y_train_step_1}
    y_pred = {'l1': forecaster.regressors_[1].predict(X_train_step_1)}
    for i in range(1, 20):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)
        for v in forecaster.out_sample_residuals_by_bin_['l1'].values():
            assert len(v) == min(5_000, 500 * i)


def test_set_out_sample_residuals_when_there_are_no_residuals_for_some_bins():
    """
    Test that set_out_sample_residuals works when there are no residuals for some bins.
    """
    rng = np.random.default_rng(12345)
    series_fit = pd.DataFrame(
        data = {
            'l1': rng.normal(loc=10, scale=1, size=100),
            'l2': rng.normal(loc=10, scale=1, size=100)
        },
        index = pd.date_range(start="01-01-2000", periods=100, freq="h"),
    )

    forecaster = ForecasterDirectMultiVariate(
                     regressor          = LinearRegression(),
                     level              = 'l1',
                     steps              = 1,
                     lags               = 5,
                     transformer_series = None,
                     binner_kwargs      = {'n_bins': 3}
                 )
    forecaster.fit(series=series_fit)
    y = series_fit['l1'].copy()
    y_pred = {'l1': y.loc[y > 10]}
    y_true = {'l1': y_pred['l1'] + rng.normal(loc=0, scale=1, size=len(y_pred['l1']))}

    warn_msg = re.escape(
        f"The following bins of level 'l1' have no out of sample residuals: [0]. "
        f"No predicted values fall in the interval "
        f"[{forecaster.binner_intervals_['l1'][0]}]. "
        f"Empty bins will be filled with a random sample of residuals."
    )
    with pytest.warns(ResidualsUsageWarning, match=warn_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)

    assert len(forecaster.out_sample_residuals_by_bin_['l1'][0]) == len(y_pred['l1'])


def test_forecaster_set_out_sample_residuals_when_transformer_y_and_differentiation():
    """
    Test set_out_sample_residuals when forecaster has transformer_y and differentiation.
    Stored should equivalent to residuals calculated manually if transformer_y and
    differentiation are applied to `y_true` and `y_pred` before calculating residuals.
    """
    rng = np.random.default_rng(12345)
    series_train = pd.DataFrame({
        'l1': pd.Series(
            rng.normal(loc=0, scale=1, size=100),
            index = pd.date_range(start='1-1-2018', periods=100, freq='D')
        ),
        'l2': pd.Series(
            rng.normal(loc=0, scale=1, size=100),
            index = pd.date_range(start='1-1-2018', periods=100, freq='D')
        )
    })
    y_true  = {
        'l1': rng.normal(loc=0, scale=1, size=5)
    }
    y_pred = {
        'l1': rng.normal(loc=0, scale=1, size=5)
    }
    forecaster = ForecasterDirectMultiVariate(
                     regressor          = LinearRegression(),
                     level              = 'l1',
                     steps              = 2,
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     differentiation    = 1
                 )
    forecaster.fit(series=series_train)
    forecaster.set_out_sample_residuals(
        y_true = y_true,
        y_pred = y_pred
    )

    y_true['l1'] = forecaster.transformer_series_['l1'].transform(y_true['l1'].reshape(-1, 1)).flatten()
    y_pred['l1'] = forecaster.transformer_series_['l1'].transform(y_pred['l1'].reshape(-1, 1)).flatten()
    y_true['l1'] = forecaster.differentiator_['l1'].transform(y_true['l1'])[forecaster.differentiation:]
    y_pred['l1'] = forecaster.differentiator_['l1'].transform(y_pred['l1'])[forecaster.differentiation:]
    residuals = {'l1': y_true['l1'] - y_pred['l1']}

    assert forecaster.out_sample_residuals_.keys() == residuals.keys()
    for key in residuals.keys():
        np.testing.assert_array_almost_equal(forecaster.out_sample_residuals_[key], residuals[key])



# Unit test __init__ ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirectMultiVariate


def test_init_TypeError_when_level_is_not_a_str():
    """
    Test TypeError is raised when level is not a str.
    """
    level = 5
    err_msg = re.escape(f"`level` argument must be a str. Got {type(level)}.")
    with pytest.raises(TypeError, match = err_msg):
        ForecasterDirectMultiVariate(LinearRegression(), level=level, lags=2, steps=3)


def test_init_TypeError_when_steps_is_not_int():
    """
    Test TypeError is raised when steps is not an int.
    """
    steps = 'not_valid_type'
    err_msg = re.escape(
        f"`steps` argument must be an int greater than or equal to 1. "
        f"Got {type(steps)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        ForecasterDirectMultiVariate(LinearRegression(), level='l1', lags=2, steps=steps)


def test_init_ValueError_when_steps_is_less_than_1():
    """
    Test ValueError is raised when steps is less than 1.
    """
    steps = 0
    err_msg = re.escape(f"`steps` argument must be greater than or equal to 1. Got {steps}.")
    with pytest.raises(ValueError, match = err_msg):
        ForecasterDirectMultiVariate(LinearRegression(), level='l1', lags=2, steps=steps)


@pytest.mark.parametrize("lags", 
                         [None, {'l1': None, 'l2': None}, {'l1': [], 'l2': []}], 
                         ids = lambda lags: f'lags: {lags}')
def test_init_ValueError_when_no_lags_or_window_features(lags):
    """
    Test ValueError is raised when no lags or window_features are passed.
    """
    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterDirectMultiVariate(
            regressor       = LinearRegression(),
            level           = 'l1',
            steps           = 3,
            lags            = lags,
            window_features = None
        )


@pytest.mark.parametrize("lags, window_features, expected", 
                         [(5, None, 5), 
                          (None, True, 6),
                          ([], True, 6),
                          (5, True, 6)], 
                         ids = lambda dt: f'lags, window_features, expected: {dt}')
def test_init_window_size_correctly_stored(lags, window_features, expected):
    """
    Test window_size is correctly stored when lags or window_features are passed.
    """
    if window_features:
        window_features = RollingFeatures(
            stats=['ratio_min_max', 'median'], window_sizes=[5, 6]
        )

    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LinearRegression(),
                     level           = 'l1',
                     steps           = 3,
                     lags            = lags,
                     window_features = window_features
                 )
    
    assert forecaster.window_size == expected
    if lags:
        np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
        assert forecaster.lags_names == [f'lag_{i}' for i in range(1, lags + 1)]
        assert forecaster.max_lag == lags
    else:
        assert forecaster.lags is None
        assert forecaster.lags_names is None
        assert forecaster.max_lag is None
    if window_features:
        assert forecaster.window_features_names == ['roll_ratio_min_max_5', 'roll_median_6']
        assert forecaster.window_features_class_names == ['RollingFeatures']
    else:
        assert forecaster.window_features_names is None
        assert forecaster.window_features_class_names is None


@pytest.mark.parametrize("lags, expected",
    [({'l1': 3, 'l2': 5}, {'l1': np.array([1, 2, 3]), 'l2': np.array([1, 2, 3, 4, 5])}), 
     ({'l1': None, 'l2': 5}, {'l1': None, 'l2': np.array([1, 2, 3, 4, 5])}),
     ({'l1': [], 'l2': 5}, {'l1': None, 'l2': np.array([1, 2, 3, 4, 5])})],
    ids = lambda value: f'lags: {value}')
def test_init_max_lag_stored_when_dict(lags, expected):
    """
    Test lags are correctly stored when passed as a dictionary.
    """

    forecaster = ForecasterDirectMultiVariate(
        LinearRegression(), level='l1', lags=lags, steps=2
    )
    
    for k in forecaster.lags:
        if forecaster.lags[k] is None:
            assert forecaster.lags[k] is None
            assert forecaster.lags_names[k] is None
        else:
            np.testing.assert_array_almost_equal(forecaster.lags[k], expected[k])
            assert forecaster.lags_names[k] == [f'{k}_lag_{i}' for i in range(1, lags[k] + 1)]
    assert forecaster.max_lag == 5
    assert forecaster.window_size == 5


@pytest.mark.parametrize("lags", 
                         [{'l1': None, 'l2': None}, {'l1': [], 'l2': []}],
                         ids = lambda lags: f'lags: {lags}')
def test_init_when_lags_dict_with_all_None(lags):
    """
    Test lags dict are correctly stored when all values are None.
    """
    window_features = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=[5, 6]
    )

    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LinearRegression(),
                     level           = 'l1',
                     steps           = 3,
                     lags            = lags,
                     window_features = window_features
                 )
    
    assert forecaster.window_size == 6
    assert forecaster.lags == {'l1': None, 'l2': None}
    assert forecaster.lags_names == {'l1': None, 'l2': None}
    assert forecaster.max_lag is None
    assert forecaster.window_features_names == ['roll_ratio_min_max_5', 'roll_median_6']
    assert forecaster.window_features_class_names == ['RollingFeatures']


@pytest.mark.parametrize("dif", 
                         [0, 0.5, 1.5, 'not_int'], 
                         ids = lambda dif: f'differentiation: {dif}')
def test_init_ValueError_when_differentiation_argument_is_not_int_or_greater_than_0(dif):
    """
    Test ValueError is raised when differentiation is not an int or greater than 0.
    """
    err_msg = re.escape(
        (f"Argument `differentiation` must be an integer equal to or "
         f"greater than 1. Got {dif}.")
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterDirectMultiVariate(
            regressor       = LinearRegression(),
            level           = 'l1',
            steps           = 3,
            lags            = 5,
            differentiation = dif
        )


@pytest.mark.parametrize("dif", 
                         [1, 2], 
                         ids = lambda dif: f'differentiation: {dif}')
def test_init_window_size_is_increased_when_differentiation(dif):
    """
    Test window_size is increased when including differentiation.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LinearRegression(),
                     level           = 'l1',
                     steps           = 3,
                     lags            = 5,
                     differentiation = dif
                 )
    
    assert forecaster.differentiation == dif
    assert forecaster.differentiation_max == dif
    assert forecaster.window_size == 5 + dif


@pytest.mark.parametrize("n_jobs", 
                         [1.0, 'not_int_auto'], 
                         ids = lambda value: f'n_jobs: {value}')
def test_init_TypeError_when_n_jobs_not_int_or_auto(n_jobs):
    """
    Test TypeError is raised in when n_jobs is not an integer or 'auto'.
    """
    err_msg = re.escape(f"`n_jobs` must be an integer or `'auto'`. Got {type(n_jobs)}.")
    with pytest.raises(TypeError, match = err_msg):
        ForecasterDirectMultiVariate(
            LinearRegression(), level='l1', steps=2, lags=2, n_jobs=n_jobs
        )



# Unit test reshape_series_long_to_dict
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from ...preprocessing import reshape_series_long_to_dict
from ....exceptions import MissingValuesWarning

# Fixtures
from .fixtures_preprocessing import values_A, values_B, values_C
from .fixtures_preprocessing import index_A, index_B, index_C
from .fixtures_preprocessing import series_long
series_long_multiindex = series_long.set_index(["series_id", "datetime"])


def test_TypeError_when_data_is_not_dataframe():
    """
    Raise TypeError if data is not a pandas DataFrame.
    """
    err_msg = "`data` must be a pandas DataFrame."
    with pytest.raises(TypeError, match=err_msg):
        reshape_series_long_to_dict(
            data='not_a_dataframe',
            series_id="series_id",
            index="datetime",
            values="values",
            freq="D",
        )


def test_reshape_series_long_to_dict_raise_value_error_when_arguments_series_id_index_values_not_provided():
    """
    Check that ValueError is raised when the input dataframe does not have MultiIndex and the
    arguments `series_id`, `index` and `values` are not provided
    """
    err_msg = (
        "Arguments `series_id`, `index`, and `values` must be "
        "specified when the input DataFrame does not have a MultiIndex. "
        "Please provide a value for each of these arguments."
    )
    with pytest.raises(ValueError, match=err_msg):
        reshape_series_long_to_dict(data=series_long, freq="D")


def test_ValueError_when_series_id_not_in_data():
    """
    Raise ValueError if series_id is not in data.
    """
    series_id = "series_id_not_in_data"

    err_msg = f"Column '{series_id}' not found in `data`."
    with pytest.raises(ValueError, match=err_msg):
        reshape_series_long_to_dict(
            data=series_long,
            series_id=series_id,
            index="datetime",
            values="values",
            freq="D",
        )


def test_ValueError_when_index_not_in_data():
    """
    Raise ValueError if index is not in data.
    """
    index = "series_id_not_in_data"

    err_msg = f"Column '{index}' not found in `data`."
    with pytest.raises(ValueError, match=err_msg):
        reshape_series_long_to_dict(
            data=series_long,
            series_id="series_id",
            index=index,
            values="values",
            freq="D",
        )


def test_ValueError_when_values_not_in_data():
    """
    Raise ValueError if values is not in data.
    """
    values = "values_not_in_data"

    err_msg = f"Column '{values}' not found in `data`."
    with pytest.raises(ValueError, match=err_msg):
        reshape_series_long_to_dict(
            data=series_long,
            series_id="series_id",
            index="datetime",
            values=values,
            freq="D",
        )


def test_MissingValuesWarning_when_series_is_incomplete():
    """
    Raise MissingValuesWarning if series is incomplete and NaN values are introduced
    after setting the index frequency.
    """
    data = pd.DataFrame({
        "series_id": ["A"] * 4 + ["B"] * 4,
        "index": pd.date_range("2020-01-01", periods=4, freq="D").tolist() * 2,
        "values": [1., 2., 3., 4.] * 2,
    })
    data = data.iloc[[0, 1, 2, 3, 4, 5, 7]]

    expected = {
        "A": pd.Series(
                np.array([1., 2., 3., 4.]), 
                index= pd.date_range("2020-01-01", periods=4, freq="D"), 
                name="A"
            ),
        "B": pd.Series(
                np.array([1., 2., np.nan, 4.]), 
                index= pd.date_range("2020-01-01", periods=4, freq="D"), 
                name="B"
            )
    }

    msg = (
        "Series 'B' is incomplete. NaNs have been introduced after setting the frequency."
    )
    with pytest.warns(MissingValuesWarning, match=msg):
        results = reshape_series_long_to_dict(
                      data=data,
                      series_id="series_id",
                      index="index",
                      values="values",
                      freq="D",
                  )

    for k in expected.keys():
        pd.testing.assert_series_equal(results[k], expected[k])


def test_check_output_reshape_series_long_to_dict():
    """
    Check output of reshape_series_long_to_dict.
    """

    expected = {
        "A": pd.Series(values_A, index=index_A, name="A"),
        "B": pd.Series(values_B, index=index_B, name="B"),
        "C": pd.Series(values_C, index=index_C, name="C"),
    }

    results = reshape_series_long_to_dict(
        data=series_long,
        series_id="series_id",
        index="datetime",
        values="values",
        freq="D",
    )

    for k in expected.keys():
        pd.testing.assert_series_equal(results[k], expected[k])


def test_check_output_reshape_series_long_to_dict_when_multiindex():
    """
    Check output of reshape_series_long_to_dict when data is a MultiIndex DataFrame.
    """

    expected = {
        "A": pd.Series(values_A, index=index_A, name="A"),
        "B": pd.Series(values_B, index=index_B, name="B"),
        "C": pd.Series(values_C, index=index_C, name="C"),
    }

    results = reshape_series_long_to_dict(
        data=series_long_multiindex,
        freq="D",
    )

    for k in expected.keys():
        pd.testing.assert_series_equal(results[k], expected[k])



from ._split import (
    TimeSeriesFold,
    OneStepAheadFold
)
from ._validation import (
    backtesting_forecaster,
    backtesting_forecaster_multiseries,
    backtesting_sarimax
)
from ._search import (
    grid_search_forecaster,
    random_search_forecaster,
    bayesian_search_forecaster,
    grid_search_forecaster_multiseries,
    random_search_forecaster_multiseries,
    bayesian_search_forecaster_multiseries,
    grid_search_sarimax,
    random_search_sarimax
)


# Unit test exog_to_direct
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.utils import exog_to_direct
from pandas.api.types import CategoricalDtype


def test_exog_to_direct_TypeError_when_exog_not_pandas_series_or_dataframe():
    """
    Test TypeError is raised when exog is not a pandas Series or DataFrame.
    """
    exog = np.arange(7)

    err_msg = re.escape(f"`exog` must be a pandas Series or DataFrame. Got {type(exog)}.")
    with pytest.raises(TypeError, match = err_msg):
        exog_to_direct(exog=exog, steps=2)


@pytest.mark.parametrize("exog                                               , dtype", 
                         [(pd.Series(np.arange(10), name='exog', dtype=int)  , int),
                          (pd.Series(np.arange(10), name='exog', dtype=float), float),
                          (pd.DataFrame(pd.Series(np.arange(10), dtype=int)  , columns=['exog']), int),
                          (pd.DataFrame(pd.Series(np.arange(10), dtype=float), columns=['exog']), float)], 
                         ids = lambda exog: f'exog: {exog}')
def test_exog_to_direct_when_steps_2_exog_1d_float_int(exog, dtype):
    """
    Test exog_to_direct results when using steps 2 and exog is a  
    1d pandas Series or 1 column DataFrame of ints or floats.
    """
    results = exog_to_direct(exog=exog, steps=2)
    expected = (
        pd.DataFrame(
            data = np.array([[0, 1],
                            [1, 2],
                            [2, 3],
                            [3, 4],
                            [4, 5],
                            [5, 6],
                            [6, 7],
                            [7, 8],
                            [8, 9]]),
            index   = pd.RangeIndex(start=1, stop=10, step=1),
            columns = ['exog_step_1', 'exog_step_2']
        ).astype({'exog_step_1': dtype, 'exog_step_2': dtype}
        ),
        ['exog_step_1', 'exog_step_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


@pytest.mark.parametrize("exog                                              , dtype", 
                         [(pd.Series(np.arange(5), name='exog', dtype=int)  , int),
                          (pd.Series(np.arange(5), name='exog', dtype=float), float),
                          (pd.DataFrame(pd.Series(np.arange(5), dtype=int)  , columns=['exog']), int),
                          (pd.DataFrame(pd.Series(np.arange(5), dtype=float), columns=['exog']), float)], 
                         ids = lambda exog: f'exog: {exog}')
def test_exog_to_direct_when_steps_series_len_exog_1d_float_int(exog, dtype):
    """
    Test exog_to_direct results when using steps equal to series len (5) and 
    exog is a 1d pandas Series or 1 column DataFrame of ints or floats.
    """
    results = exog_to_direct(exog=exog, steps=len(exog))
    expected = (
        pd.DataFrame(
            data    = np.array([[0, 1, 2, 3, 4]]),
            index   = pd.RangeIndex(start=4, stop=5, step=1),
            columns = ['exog_step_1', 'exog_step_2', 'exog_step_3',
                       'exog_step_4', 'exog_step_5']
        ).astype({'exog_step_1': dtype, 'exog_step_2': dtype,
                  'exog_step_3': dtype, 'exog_step_4': dtype,
                  'exog_step_5': dtype}
        ),
        ['exog_step_1', 'exog_step_2', 'exog_step_3', 'exog_step_4', 'exog_step_5']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_exog_to_direct_when_steps_2_exog_DataFrame_2d_float_int(dtype):
    """
    Test exog_to_direct results when using steps 2 and exog is a  
    pandas DataFrame with 2 columns of floats or ints.
    """
    exog = pd.DataFrame({'exog_1': np.arange(100, 110, dtype=dtype),
                         'exog_2': np.arange(1000, 1010, dtype=dtype)})
    results = exog_to_direct(exog=exog, steps=2)
    expected = (
        pd.DataFrame(
            data = np.array([[100, 1000, 101, 1001],
                            [101, 1001, 102, 1002],
                            [102, 1002, 103, 1003],
                            [103, 1003, 104, 1004],
                            [104, 1004, 105, 1005],
                            [105, 1005, 106, 1006],
                            [106, 1006, 107, 1007],
                            [107, 1007, 108, 1008],
                            [108, 1008, 109, 1009]]),
            index   = pd.RangeIndex(start=1, stop=10, step=1),
            columns = ['exog_1_step_1', 'exog_2_step_1', 
                       'exog_1_step_2', 'exog_2_step_2']
        ).astype({'exog_1_step_1': dtype, 'exog_2_step_1': dtype,
                  'exog_1_step_2': dtype, 'exog_2_step_2': dtype}
        ),
        ['exog_1_step_1', 'exog_2_step_1', 'exog_1_step_2', 'exog_2_step_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


def test_exog_to_direct_when_steps_2_exog_1d_category():
    """
    Test exog_to_direct results when using steps 2 and exog is a  
    1d pandas Series or 1 column DataFrame of category.
    """
    exog = pd.Series(range(10), name='exog', dtype='category')
    results = exog_to_direct(exog=exog, steps=2)
    expected = (
        pd.DataFrame(
            {'exog_step_1': pd.Categorical(range(9), categories=range(10)),
             'exog_step_2': pd.Categorical(range(1, 10), categories=range(10))},
            index = pd.RangeIndex(start=1, stop=10, step=1)
        ),
        ['exog_step_1', 'exog_step_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


def test_exog_to_direct_when_steps_2_exog_DataFrame_2d_category():
    """
    Test exog_to_direct results when using steps 2 and exog is a  
    pandas DataFrame with 2 columns of category.
    """
    exog = pd.DataFrame({'exog_1': pd.Categorical(range(10)),
                         'exog_2': pd.Categorical(range(100, 110))})
    results = exog_to_direct(exog=exog, steps=2)
    expected = (
        pd.DataFrame({
            'exog_1_step_1': pd.Categorical(range(9), categories=range(10)),
            'exog_2_step_1': pd.Categorical(range(100, 109), categories=range(100, 110)),
            'exog_1_step_2': pd.Categorical(range(1, 10), categories=range(10)),
            'exog_2_step_2': pd.Categorical(range(101, 110), categories=range(100, 110))},
            index = pd.RangeIndex(start=1, stop=10, step=1)
        ),
        ['exog_1_step_1', 'exog_2_step_1', 
         'exog_1_step_2', 'exog_2_step_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


def test_exog_to_direct_when_steps_2_exog_DataFrame_3d_float_int_category():
    """
    Test exog_to_direct results when using steps 2 and exog is a  
    pandas DataFrame with 3 columns of float, int, category.
    """
    exog = pd.DataFrame({'exog_1': pd.Series(np.arange(100, 110), dtype='float64'),
                         'exog_2': pd.Series(np.arange(1000, 1010), dtype='int64'),
                         'exog_3': pd.Categorical(range(100, 110))})
    results = exog_to_direct(exog=exog, steps=2)
    expected = (
        pd.DataFrame(
            data = np.array([[100., 1000., 100., 101., 1001., 101.],
                            [101., 1001., 101., 102., 1002., 102.],
                            [102., 1002., 102., 103., 1003., 103.],
                            [103., 1003., 103., 104., 1004., 104.],
                            [104., 1004., 104., 105., 1005., 105.],
                            [105., 1005., 105., 106., 1006., 106.],
                            [106., 1006., 106., 107., 1007., 107.],
                            [107., 1007., 107., 108., 1008., 108.],
                            [108., 1008., 108., 109., 1009., 109.]]),
            index   = pd.RangeIndex(start=1, stop=10, step=1),
            columns = ['exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2']
        ).astype(
            {'exog_1_step_1': 'float64',
             'exog_2_step_1': 'int64',
             'exog_3_step_1': 'int64',
             'exog_1_step_2': 'float64',
             'exog_2_step_2': 'int64',
             'exog_3_step_2': 'int64'}
        ).astype(
            {'exog_3_step_1': CategoricalDtype(categories=[100, 101, 102, 103, 104, 105, 106, 107, 108, 109], ordered=False),
             'exog_3_step_2': CategoricalDtype(categories=[100, 101, 102, 103, 104, 105, 106, 107, 108, 109], ordered=False)}
        ),
        ['exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


def test_exog_to_direct_when_steps_len_exog_exog_DataFrame_3d_float_int_category():
    """
    Test exog_to_direct results when using steps len exog (5) and exog is a  
    pandas DataFrame with 3 columns of float, int, category.
    """
    exog = pd.DataFrame({'exog_1': pd.Series(np.arange(100, 103), dtype='float64'),
                         'exog_2': pd.Series(np.arange(1000, 1003), dtype='int64'),
                         'exog_3': pd.Categorical(range(100, 103))})
    results = exog_to_direct(exog=exog, steps=len(exog))
    expected = (
        pd.DataFrame(
            data = np.array([[100., 1000., 100., 101., 1001., 101., 102., 1002., 102.]]),
            index   = pd.RangeIndex(start=2, stop=3, step=1),
            columns = ['exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2', 
                       'exog_1_step_3', 'exog_2_step_3', 'exog_3_step_3']
        ).astype(
            {'exog_1_step_1': 'float64',
             'exog_2_step_1': 'int64',
             'exog_3_step_1': 'int64',
             'exog_1_step_2': 'float64',
             'exog_2_step_2': 'int64',
             'exog_3_step_2': 'int64',
             'exog_1_step_3': 'float64',
             'exog_2_step_3': 'int64',
             'exog_3_step_3': 'int64'}
        ).astype(
            {'exog_3_step_1': CategoricalDtype(categories=[100, 101, 102], ordered=False),
             'exog_3_step_2': CategoricalDtype(categories=[100, 101, 102], ordered=False),
             'exog_3_step_3': CategoricalDtype(categories=[100, 101, 102], ordered=False)}
        ),
        ['exog_1_step_1', 'exog_2_step_1', 'exog_3_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 'exog_3_step_2', 
         'exog_1_step_3', 'exog_2_step_3', 'exog_3_step_3']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]


@pytest.mark.parametrize("dtype", 
                         [float, int], 
                         ids = lambda dt: f'dtype: {dt}')
def test_exog_to_direct_when_steps_3_exog_DataFrame_2d_float_int(dtype):
    """
    Test exog_to_direct results when using steps 3 and exog is a  
    pandas DataFrame with 2 columns of floats or ints.
    """
    exog = pd.DataFrame({'exog_1': np.arange(100, 110, dtype=dtype),
                         'exog_2': np.arange(1000, 1010, dtype=dtype)})
    results = exog_to_direct(exog=exog, steps=3)
    expected = (
        pd.DataFrame(
            data = np.array([[100, 1000, 101, 1001, 102, 1002],
                            [101, 1001, 102, 1002, 103, 1003],
                            [102, 1002, 103, 1003, 104, 1004],
                            [103, 1003, 104, 1004, 105, 1005],
                            [104, 1004, 105, 1005, 106, 1006],
                            [105, 1005, 106, 1006, 107, 1007],
                            [106, 1006, 107, 1007, 108, 1008],
                            [107, 1007, 108, 1008, 109, 1009]]),
            index   = pd.RangeIndex(start=2, stop=10, step=1),
            columns = ['exog_1_step_1', 'exog_2_step_1', 
                       'exog_1_step_2', 'exog_2_step_2', 
                       'exog_1_step_3', 'exog_2_step_3']
        ).astype({'exog_1_step_1': dtype, 'exog_2_step_1': dtype, 
                  'exog_1_step_2': dtype, 'exog_2_step_2': dtype, 
                  'exog_1_step_3': dtype, 'exog_2_step_3': dtype}
        ),
        ['exog_1_step_1', 'exog_2_step_1', 
         'exog_1_step_2', 'exog_2_step_2', 
         'exog_1_step_3', 'exog_2_step_3']
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert results[1] == expected[1]



# Unit test prepare_steps_direct
# ==============================================================================
import re
import pytest
from skforecast.utils import prepare_steps_direct


@pytest.mark.parametrize("steps", [[1, 2.0, 3], [1, 4.]], 
                         ids=lambda steps: f'steps: {steps}')
def test_TypeError_prepare_steps_direct_when_steps_list_contain_floats(steps):
    """
    Test TypeError when steps list contain floats.
    """

    err_msg = re.escape(
        (f"`steps` argument must be an int, a list of ints or `None`. "
         f"Got {type(steps)}.")
    )
    with pytest.raises(TypeError, match = err_msg):
        prepare_steps_direct(max_step=5, steps=steps)


@pytest.mark.parametrize("steps, max_step, expected_steps", 
                         [(4, 5, [1, 2, 3, 4]),
                          (None, 5, [1, 2, 3, 4, 5]),
                          (None, [1, 3], [1, 3]),
                          ([1, 3], 5, [1, 3])], 
                         ids=lambda param: f'steps: {param}')
def test_output_prepare_steps_direct(steps, max_step, expected_steps):
    """
    Test output prepare_steps_direct for different inputs.
    """
    steps = prepare_steps_direct(
                max_step = max_step, 
                steps    = steps
            )
    
    assert steps == expected_steps


# Unit test set_params ForecasterSarimax
# ==============================================================================
from skforecast.sarimax import Sarimax
from skforecast.recursive import ForecasterSarimax


def test_ForecasterSarimax_set_params_skforecast():
    """
    Test set_params() method skforecast.
    """
    forecaster = ForecasterSarimax(regressor = Sarimax(order=(1, 1, 1)))
    new_params = {'order': (2, 2, 2), 'seasonal_order': (1, 1, 1, 2)}
    forecaster.set_params(new_params)
    results = forecaster.regressor.get_params()

    expected = {
        'order': (2, 2, 2),
        'seasonal_order': (1, 1, 1, 2),
        'trend': None,
        'measurement_error': False,
        'time_varying_regression': False,
        'mle_regression': True,
        'simple_differencing': False,
        'enforce_stationarity': True,
        'enforce_invertibility': True,
        'hamilton_representation': False,
        'concentrate_scale': False,
        'trend_offset': 1,
        'use_exact_diffuse': False,
        'dates': None,
        'freq': None,
        'missing': 'none',
        'validate_specification': True,
        'method': 'lbfgs',
        'maxiter': 50,
        'start_params': None,
        'disp': False,
        'sm_init_kwargs': {},
        'sm_fit_kwargs': {},
        'sm_predict_kwargs': {}
    }

    assert results == expected



# Unit test predict method using PyTorch backend
# ==============================================================================
import os
import numpy as np
import pandas as pd
os.environ["KERAS_BACKEND"] = "torch"
import keras
from skforecast.deep_learning.utils import create_and_compile_model
from skforecast.deep_learning import ForecasterRnn

series = pd.DataFrame(
    {
        "1": np.arange(50),
        "2": np.arange(50),
        "3": np.arange(50),
    },
    index=pd.date_range("2020-01-01", periods=50, freq="D")
)

exog = pd.DataFrame(
    {
        "exog1": np.arange(50),
        "exog2": np.arange(50),
    },
    index=pd.date_range("2020-01-01", periods=50, freq="D")
)

exog_pred = pd.DataFrame(
    {
        "exog1": np.arange(50, 60),
        "exog2": np.arange(50, 60),
    },
    index=pd.date_range("2020-02-20", periods=10, freq="D")
)

model = create_and_compile_model(
            series=series, 
            levels=["1", "2"],    
            lags=3,           
            steps=4,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )

model_exog = create_and_compile_model(
            series=series, 
            exog=exog,
            levels=["1", "2", "3"],    
            lags=10,           
            steps=8,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )


def test_predict_3_steps_ahead():
    """
    Test case for predicting 3 steps ahead
    """
    forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
    forecaster.fit(series=series)
    predictions = forecaster.predict(steps=3)

    assert predictions.shape == (6, 2)


def test_predict_specific_levels():
    """
    Test case for predicting with specific levels
    """
    forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
    forecaster.fit(series=series)
    predictions = forecaster.predict(steps=None, levels=["1"])

    assert predictions.shape == (4, 2)


def test_predict_exog():
    """
    Test case for predicting with exogenous variables
    """
    forecaster = ForecasterRnn(model_exog, levels=["1", "2", "3"], lags=10)
    forecaster.fit(series=series, exog=exog)
    predictions = forecaster.predict(steps=None, exog=exog_pred)

    assert predictions.shape == (24, 2)


def test_predict_specific_levels_with_exog():
    """
    Test case for predicting with specific levels
    """
    forecaster = ForecasterRnn(model_exog, levels=["1", "2", "3"], lags=10)
    forecaster.fit(series=series, exog=exog)
    predictions = forecaster.predict(steps=5, exog=exog_pred, levels=["1", "2"])

    assert predictions.shape == (10, 2)



# Unit test select_features_multiseries
# ==============================================================================
import re
import pytest
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.feature_selection import RFE
from sklearn.preprocessing import StandardScaler
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirectMultiVariate
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.feature_selection import select_features_multiseries

# Fixtures
from .fixtures_feature_selection import (
    series_wide_range,
    series_dict_range,
    exog_multiseries as exog
)


def test_TypeError_select_features_multiseries_raise_when_forecaster_is_not_supported():
    """
    Test TypeError is raised in select_features_multiseries when forecaster is 
    not supported.
    """
    
    err_msg = re.escape(
        "`forecaster` must be one of the following classes: "
        "['ForecasterRecursiveMultiSeries', 'ForecasterDirectMultiVariate']."
    )
    with pytest.raises(TypeError, match = err_msg):
        select_features_multiseries(
            selector   = object(),
            forecaster = object(),
            series     = object(),
            exog       = object(),
        )


@pytest.mark.parametrize("select_only", 
                         ['not_exog_or_autoreg', 1, False], 
                         ids=lambda so: f'select_only: {so}')
def test_ValueError_select_features_multiseries_raise_when_select_only_is_not_autoreg_exog_None(select_only):
    """
    Test ValueError is raised in select_features_multiseries when `select_only` 
    is not 'autoreg', 'exog' or None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    err_msg = re.escape(
        "`select_only` must be one of the following values: 'autoreg', 'exog', None."
    )
    with pytest.raises(ValueError, match = err_msg):
        select_features_multiseries(
            selector    = selector,
            forecaster  = forecaster,
            series      = object(),
            exog        = object(),
            select_only = select_only,
        )


@pytest.mark.parametrize("subsample", 
                         [-1, -0.5, 0, 0., 1.1, 2], 
                         ids=lambda ss: f'subsample: {ss}')
def test_ValueError_select_features_multiseries_raise_when_subsample_is_not_greater_0_less_equal_1(subsample):
    """
    Test ValueError is raised in select_features_multiseries when `subsample` 
    is not in (0, 1].
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 5,
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)
    err_msg = re.escape(
        "`subsample` must be a number greater than 0 and less than or equal to 1."
    )
    with pytest.raises(ValueError, match = err_msg):
        select_features_multiseries(
            selector   = selector,
            forecaster = forecaster,
            series     = object(),
            exog       = object(),
            subsample  = subsample,
        )


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_exog_regressor():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is 'exog' and regressor is passed to the selector instead
    of forecaster.regressor.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 5,
                     encoding  = 'onehot'
                 )
    selector = RFE(estimator=LinearRegression(), n_features_to_select=2)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_dict_range,
        exog        = exog,
        select_only = 'exog',
        verbose     = False,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_exog_ForecasterRecursiveMultiSeries_no_window_features():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterRecursiveMultiSeries and 
    no window features are included.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 5,
                     encoding  = 'ordinal'
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=2)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_dict_range,
        exog        = exog,
        select_only = 'exog',
        verbose     = False,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_exog_ForecasterRecursiveMultiSeries_window_features():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterRecursiveMultiSeries and 
    window features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor       = LinearRegression(),
                     lags            = 5,
                     window_features = roll_features,
                     encoding        = 'ordinal'
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=2)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_dict_range,
        exog        = exog,
        select_only = 'exog',
        verbose     = False,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == ['roll_mean_3', 'roll_std_5']
    assert selected_exog == ['exog1', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_autoreg_ForecasterRecursiveMultiSeries_no_window_features():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is 'autoreg'. Forecaster is ForecasterRecursiveMultiSeries and 
    no window features are included.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 5,
                     encoding  = 'ordinal'
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=2)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_dict_range,
        exog        = exog,
        select_only = 'autoreg',
        verbose     = False,
    )

    assert selected_lags == [4, 5]
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog2', 'exog3', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_autoreg_ForecasterRecursiveMultiSeries_window_features():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is 'autoreg'. Forecaster is ForecasterRecursiveMultiSeries and 
    window features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor       = Ridge(alpha=0.1, random_state=123),
                     lags            = 5,
                     window_features = roll_features,
                     encoding        = 'ordinal'
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=4)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_dict_range,
        exog        = exog,
        select_only = 'autoreg',
        subsample   = 0.9,
        verbose     = False,
    )

    assert selected_lags == [3, 4, 5]
    assert selected_window_features == ['roll_std_5']
    assert selected_exog == ['exog1', 'exog2', 'exog3', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_None_no_window_features():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is None. Forecaster is ForecasterRecursiveMultiSeries and 
    no window features are included.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     encoding           = 'onehot',
                     transformer_series = StandardScaler()
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=2)

    warn_msg = re.escape(
        "No autoregressive features have been selected. Since a Forecaster "
        "cannot be created without them, be sure to include at least one "
        "using the `force_inclusion` parameter."
    )
    with pytest.warns(UserWarning, match = warn_msg):
        selected_lags, selected_window_features, selected_exog = select_features_multiseries(
            selector    = selector,
            forecaster  = forecaster,
            series      = series_dict_range,
            exog        = exog,
            select_only = None,
            verbose     = False,
        )

    assert selected_lags == []
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_and_select_only_is_None_window_features():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    and select_only is None. Forecaster is ForecasterRecursiveMultiSeries and 
    window features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     window_features    = roll_features,
                     encoding           = 'onehot',
                     transformer_series = StandardScaler()
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_dict_range,
        exog        = exog,
        select_only = None,
        verbose     = False,
    )

    assert selected_lags == []
    assert selected_window_features == ['roll_std_5']
    assert selected_exog == ['exog1', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_select_only_exog_is_True_and_force_inclusion_is_regex():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    select_only_exog is True and force_inclusion is regex.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor       = LinearRegression(),
                     lags            = 5,
                     window_features = roll_features,
                     encoding        = 'ordinal'
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector        = selector,
        forecaster      = forecaster,
        series          = series_dict_range,
        exog            = exog,
        select_only     = 'exog',
        force_inclusion = "^exog_3",
        verbose         = False,
    )

    assert selected_lags == [1, 2, 3, 4, 5]
    assert selected_window_features == ['roll_mean_3', 'roll_std_5']
    assert selected_exog == ['exog1', 'exog3', 'exog4']


def test_select_features_multiseries_when_selector_is_RFE_select_only_exog_is_False_and_force_inclusion_is_list():
    """
    Test that select_features_multiseries returns the expected values when selector is RFE
    select_only_exog is False and force_inclusion is list.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 5,
                     encoding  = 'onehot',
                     transformer_series = StandardScaler()
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector        = selector,
        forecaster      = forecaster,
        series          = series_dict_range,
        exog            = exog,
        select_only     = None,
        force_inclusion = ['lag_1'],
        verbose         = True,
    )

    assert selected_lags == [1, 4]
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog4']


@pytest.mark.parametrize("lags", 
                         [{'l1': None, 'l2': 5}, {'l1': [], 'l2': 5}],
                         ids = lambda lags: f'lags: {lags}')
def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterDirectMultiVariate_lags_dict(lags):
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterDirectMultiVariate 
    and lags is a dictionary.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = LinearRegression(),
                     level     = 'l1',
                     steps     = 3,
                     lags      = lags
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_wide_range,
        exog        = exog,
        select_only = 'autoreg',
        verbose     = False,
    )

    assert selected_lags == {'l1': [], 'l2': [2, 3, 4]}
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog2', 'exog3', 'exog4']


def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterDirectMultiVariate_no_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterDirectMultiVariate and 
    no window features are included.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = LinearRegression(),
                     level     = 'l1',
                     steps     = 3,
                     lags      = 5
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_wide_range,
        exog        = exog,
        select_only = 'autoreg',
        verbose     = False,
    )

    assert selected_lags == {'l1': [3, 5], 'l2': [3]}
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog2', 'exog3', 'exog4']


def test_select_features_when_selector_is_RFE_select_only_is_exog_ForecasterDirectMultiVariate_window_features():
    """
    Test that select_features returns the expected values when selector is RFE
    and select_only is 'exog'. Forecaster is ForecasterDirectMultiVariate and 
    window features are included.
    """
    roll_features = RollingFeatures(
                        stats=['mean', 'std'],
                        window_sizes=[3, 5],
                    )
    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LinearRegression(),
                     level           = 'l1',
                     steps           = 3,
                     lags            = 5,
                     window_features = roll_features
                 )
    selector = RFE(estimator=forecaster.regressor, n_features_to_select=3)

    selected_lags, selected_window_features, selected_exog = select_features_multiseries(
        selector    = selector,
        forecaster  = forecaster,
        series      = series_wide_range,
        exog        = exog,
        select_only = 'autoreg',
        verbose     = False,
    )

    assert selected_lags == {'l1': [3, 5], 'l2': [3]}
    assert selected_window_features == []
    assert selected_exog == ['exog1', 'exog2', 'exog3', 'exog4']



# Unit test create_sample_weights ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.direct import ForecasterDirectMultiVariate
from sklearn.linear_model import LinearRegression


def custom_weights(index): # pragma: no cover
    """
    Return 0 if index is between '2022-01-10', '2022-01-12', 1 otherwise.
    """
    weights = np.where(
                  (index >= '2022-01-10') & (index <= '2022-01-12'),
                   0,
                   1
              )
    
    return weights


def custom_weights_nan(index): # pragma: no cover
    """
    Return np.nan f index is between '2022-01-10', '2022-01-12', 1 otherwise.
    """
    weights = np.where(
                  (index >= '2022-01-10') & (index <= '2022-01-12'),
                   np.nan,
                   1
              )
    
    return weights


def custom_weights_negative(index): # pragma: no cover
    """
    Return -1 f index is between '2022-01-10', '2022-01-12', 1 otherwise.
    """
    weights = np.where(
                  (index >= '2022-01-10') & (index <= '2022-01-12'),
                   -1,
                   1
              )
    
    return weights


def custom_weights_zeros(index): # pragma: no cover
    """
    Return 0 for all elements in index
    """
    weights = np.zeros_like(index, dtype=int)

    return weights


series = pd.DataFrame(
             data = np.array([[0.12362923, 0.51328688],
                              [0.65138268, 0.11599708],
                              [0.58142898, 0.72350895],
                              [0.72969992, 0.10305721],
                              [0.97790567, 0.20581485],
                              [0.56924731, 0.41262027],
                              [0.85369084, 0.82107767],
                              [0.75425194, 0.0107816 ],
                              [0.08167939, 0.94951918],
                              [0.00249297, 0.55583355]]),
             columns= ['series_1', 'series_2'],
             index = pd.DatetimeIndex(
                         ['2022-01-04', '2022-01-05', '2022-01-06', '2022-01-07',
                          '2022-01-08', '2022-01-09', '2022-01-10', '2022-01-11',
                          '2022-01-12', '2022-01-13'],
                         dtype='datetime64[ns]', freq='D'
                     ),
         )

# X_train for 3 lags and 3 steps
X_train = pd.DataFrame(
              data = np.array(
                          [[0.58142898, 0.65138268, 0.12362923, 0.72350895, 0.11599708, 0.51328688],
                           [0.72969992, 0.58142898, 0.65138268, 0.10305721, 0.72350895, 0.11599708],  
                           [0.97790567, 0.72969992, 0.58142898, 0.20581485, 0.10305721, 0.72350895],
                           [0.56924731, 0.97790567, 0.72969992, 0.41262027, 0.20581485, 0.10305721],
                           [0.85369084, 0.56924731, 0.97790567, 0.82107767, 0.41262027, 0.20581485]]
                     ),
               index = pd.DatetimeIndex(
                           ['2022-01-09', '2022-01-10', '2022-01-11', '2022-01-12', '2022-01-13'],
                           dtype='datetime64[ns]', freq='D'
                       ),
               columns = ['series_1_lag_1', 'series_1_lag_2', 'series_1_lag_3', 
                          'series_2_lag_1', 'series_2_lag_2', 'series_2_lag_3']
          )


def test_create_sample_weights_output():
    """
    Test sample_weights creation.
    """
    forecaster = ForecasterDirectMultiVariate(
                     level       = 'series_1',
                     lags        = 3,
                     steps       = 3,    
                     regressor   = LinearRegression(),
                     weight_func = custom_weights
                 )

    expected = np.array([1, 0, 0, 0, 1])
    results = forecaster.create_sample_weights(X_train=X_train)

    assert np.array_equal(results, expected)


def test_create_sample_weights_exceptions_when_weights_has_nan():
    """
    Test sample_weights exception when weights contains NaNs.
    """
    forecaster = ForecasterDirectMultiVariate(
                     level       = 'series_1',
                     lags        = 3,
                     steps       = 3,  
                     regressor   = LinearRegression(),
                     weight_func = custom_weights_nan
                 )

    err_msg = re.escape("The resulting `sample_weight` cannot have NaN values.")
    with pytest.raises(ValueError, match=err_msg):
        forecaster.create_sample_weights(X_train=X_train)
    

def test_create_sample_weights_exceptions_when_weights_has_negative_values():
    """
    Test sample_weights exception when sample_weight contains negative values.
    """
    forecaster = ForecasterDirectMultiVariate(
                     level       = 'series_1',
                     lags        = 3,
                     steps       = 3,     
                     regressor   = LinearRegression(),
                     weight_func = custom_weights_negative
                 )

    err_msg = re.escape("The resulting `sample_weight` cannot have negative values.")
    with pytest.raises(ValueError, match=err_msg):
        forecaster.create_sample_weights(X_train=X_train)
    

def test_create_sample_weights_exceptions_when_weights_all_zeros():
    """
    Test sample_weights exception when all weights are zeros.
    """
    forecaster = ForecasterDirectMultiVariate(
                     level       = 'series_1',
                     lags        = 3,
                     steps       = 3,    
                     regressor   = LinearRegression(),
                     weight_func = custom_weights_zeros
                 )
    
    err_msg = re.escape(
                    ("The resulting `sample_weight` cannot be normalized because "
                     "the sum of the weights is zero.")
                )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.create_sample_weights(X_train=X_train)


from .preprocessing import (
    TimeSeriesDifferentiator,
    DateTimeFeatureTransformer,
    create_datetime_features,
    reshape_series_wide_to_long,
    reshape_series_long_to_dict,
    reshape_exog_long_to_dict,
    RollingFeatures,
    QuantileBinner,
    FastOrdinalEncoder,
    ConformalIntervalCalibrator,
)


# Unit test set_fit_kwargs ForecasterSarimax
# ==============================================================================
import re
import pytest
from skforecast.sarimax import Sarimax
from skforecast.recursive import ForecasterSarimax
from skforecast.exceptions import IgnoredArgumentWarning


def test_set_fit_kwargs_skforecast():
    """
    Test set_fit_kwargs method using skforecast.
    """
    forecaster = ForecasterSarimax(
                     regressor  = Sarimax(order=(1, 0, 1))
                 )
    new_fit_kwargs = {'warning': 1}
    
    warn_msg = re.escape(
        ("When using the skforecast Sarimax model, the fit kwargs should "
         "be passed using the model parameter `sm_fit_kwargs`.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        forecaster.set_fit_kwargs(new_fit_kwargs)
    
    assert forecaster.fit_kwargs == {}



# Unit test predict_quantiles ForecasterDirectMultiVariate
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.direct import ForecasterDirectMultiVariate
from sklearn.linear_model import LinearRegression
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder

# Fixtures
from .fixtures_forecaster_direct_multivariate import series
from .fixtures_forecaster_direct_multivariate import exog
from .fixtures_forecaster_direct_multivariate import exog_predict

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer():
    """
    Test output of predict_quantiles when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor          = LinearRegression(),
                     steps              = 2,
                     level              = 'l1',
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    results = forecaster.predict_quantiles(
                  steps                   = 2,
                  exog                    = exog_predict,
                  quantiles               = [0.05, 0.55, 0.95],
                  n_boot                  = 4,
                  use_in_sample_residuals = True, 
                  use_binned_residuals    = False
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.49293687, 0.69241978, 0.78565834],
                                        [0.09125144, 0.3137743 , 0.56737647]]),
                   columns = ['q_0.05', 'q_0.55', 'q_0.95'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_quantiles when regressor is LinearRegression,
    2 steps are predicted, using out-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor          = LinearRegression(),
                     steps              = 2,
                     level              = 'l1',
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_
    results = forecaster.predict_quantiles(
                  steps                   = 2,
                  exog                    = exog_predict,
                  quantiles               = (0.05, 0.55, 0.95),
                  n_boot                  = 4,
                  use_in_sample_residuals = False, 
                  use_binned_residuals    = False
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.49293687, 0.69241978, 0.78565834],
                                        [0.09125144, 0.3137743 , 0.56737647]]),
                   columns = ['q_0.05', 'q_0.55', 'q_0.95'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))

    pd.testing.assert_frame_equal(expected, results)



# Unit test create_train_X_y ForecasterDirect
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.direct import ForecasterDirect


def test_create_train_X_y_output_when_y_is_series_10_steps_1_and_exog_is_series_of_float():
    """
    Test the output of create_train_X_y when y=pd.Series(np.arange(10)), steps=1 
    and exog is a pandas Series of floats.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=float)

    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=1)
    results = forecaster.create_train_X_y(y=y, exog=exog)

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105.],
                             [5., 4., 3., 2., 1., 106.],
                             [6., 5., 4., 3., 2., 107.],
                             [7., 6., 5., 4., 3., 108.],
                             [8., 7., 6., 5., 4., 109.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'exog_step_1']
        ).astype({'exog_step_1': float}),
        {1: pd.Series(
                data  = np.array([5., 6., 7., 8., 9.], dtype=float), 
                index = pd.RangeIndex(start=5, stop=10, step=1),
                name  = "y_step_1"
            )
        }
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], dict)
    assert all(isinstance(x, pd.Series) for x in results[1].values())
    assert results[1].keys() == expected[1].keys()
    for key in expected[1]: 
        pd.testing.assert_series_equal(results[1][key], expected[1][key]) 



# Unit test _create_window_features ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect


class WindowFeatureNoPandas:  # pragma: no cover
    def __init__(self, window_sizes, features_names):
        self.window_sizes = window_sizes
        self.features_names = features_names

    def transform_batch(self, y):
        return y

    def transform(self):
        pass


class WindowFeatureNoCorrectLength:  # pragma: no cover
    def __init__(self, window_sizes, features_names):
        self.window_sizes = window_sizes
        self.features_names = features_names

    def transform_batch(self, y):
        y = pd.DataFrame(y).iloc[-1:, :]
        return y

    def transform(self):
        pass


class WindowFeatureNoCorrectIndex:  # pragma: no cover
    def __init__(self, window_sizes, features_names):
        self.window_sizes = window_sizes
        self.features_names = features_names

    def transform_batch(self, y):
        y = pd.DataFrame(y)
        y.index = pd.RangeIndex(start=0, stop=len(y), step=1)
        return y

    def transform(self):
        pass


def test_create_window_features_TypeError_when_transform_batch_not_pandas():
    """
    Test TypeError is raised when `transform_batch` does not return 
    a pandas DataFrame.
    """
    wf = WindowFeatureNoPandas(window_sizes=5, features_names='feature_1')
    y = pd.Series(np.arange(10))
    train_index = pd.RangeIndex(start=6, stop=10, step=1)

    forecaster = ForecasterDirect(
        LinearRegression(), lags=5, steps=2, window_features=wf
    )
    err_msg = re.escape(
        ("The method `transform_batch` of WindowFeatureNoPandas "
         "must return a pandas DataFrame.")
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster._create_window_features(y=y, train_index=train_index)


def test_create_window_features_ValueError_when_transform_batch_not_correct_length():
    """
    Test ValueError is raised when `transform_batch` does not return
    a DataFrame with the correct length.
    """
    wf = WindowFeatureNoCorrectLength(window_sizes=5, features_names='feature_1')
    y = pd.Series(np.arange(10))
    train_index = pd.RangeIndex(start=6, stop=10, step=1)

    forecaster = ForecasterDirect(
        LinearRegression(), lags=5, steps=2, window_features=wf
    )
    err_msg = re.escape(
        ("The method `transform_batch` of WindowFeatureNoCorrectLength "
         "must return a DataFrame with the same number of rows as "
         "the input time series - (`window_size` + (`steps` - 1)): 4.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_window_features(y=y, train_index=train_index)


def test_create_window_features_output():
    """
    Test window features are created properly.
    """
    y = pd.Series(np.arange(10))
    train_index = pd.RangeIndex(start=6, stop=10, step=1)
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        LinearRegression(), lags=5, steps=2, window_features=rolling
    )
    results = forecaster._create_window_features(
        y=y, train_index=train_index
    )
    expected = (
        [
            np.array([[3., 3., 15.],
                      [4., 4., 21.],
                      [5., 5., 27.],
                      [6., 6., 33.]])
        ],
        ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    )

    for result, exp in zip(results[0], expected[0]):
        np.testing.assert_array_almost_equal(result, exp)
    assert results[1] == expected[1]


def test_create_window_features_output_as_pandas():
    """
    Test window features are created properly as pandas.
    """
    y_datetime = pd.Series(
        np.arange(10), pd.date_range(start='2020-01-01', periods=10)
    )
    train_index = pd.date_range(start='2020-01-08', periods=3)
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        LinearRegression(), lags=3, steps=3, window_features=rolling
    )
    results = forecaster._create_window_features(
        y=y_datetime, train_index=train_index, X_as_pandas=True
    )
    expected = (
        [
            pd.DataFrame(
                data = np.array(
                           [[4., 4., 21.],
                            [5., 5., 27.],
                            [6., 6., 33.]]),
                index = train_index,
                columns = ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
            )
        ],
        ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    )

    for result, exp in zip(results[0], expected[0]):
        np.testing.assert_array_almost_equal(result, exp)
    assert results[1] == expected[1]


def test_create_window_features_output_list():
    """
    Test window features are created properly when `window_features` is a list.
    """
    y = pd.Series(np.arange(10))
    train_index = pd.RangeIndex(start=7, stop=10, step=1)
    rolling_1 = RollingFeatures(
        stats=['mean', 'median'], window_sizes=[5, 5]
    )
    rolling_2 = RollingFeatures(
        stats='sum', window_sizes=6, features_names=['feature_2']
    )

    forecaster = ForecasterDirect(
        LinearRegression(), lags=3, steps=2, window_features=[rolling_1, rolling_2]
    )
    results = forecaster._create_window_features(y=y, train_index=train_index)
    expected = (
        [
            np.array([[4., 4.],
                      [5., 5.],
                      [6., 6.]]),
            np.array([[21.],
                      [27.],
                      [33.]])
        ],
        ['roll_mean_5', 'roll_median_5', 'feature_2']
    )

    for result, exp in zip(results[0], expected[0]):
        np.testing.assert_array_almost_equal(result, exp)
    assert results[1] == expected[1]



# Unit test set_out_sample_residuals ForecasterRnn
# ==============================================================================
import os
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
os.environ["KERAS_BACKEND"] = "torch"
import keras
from skforecast.deep_learning import create_and_compile_model
from skforecast.deep_learning import ForecasterRnn

# Fixtures
np.random.seed(123)
series = pd.DataFrame(
    {
        "l1": np.random.rand(50),
        "l2": np.random.rand(50)
    },
    index=pd.date_range(start='2000-01-01', periods=50, freq='MS')
)
exog = pd.DataFrame(
    {
        "exog_1": np.random.rand(50),
        "exog_2": np.random.rand(50)
    },
    index=pd.date_range(start='2000-01-01', periods=50, freq='MS')
)

model = create_and_compile_model(
    series=series,
    exog=exog,
    lags=3,
    steps=5,
    levels=["l1", "l2"],
    recurrent_units=64,
    dense_units=32,
)

model_no_exog = create_and_compile_model(
    series=series,
    lags=3,
    steps=5,
    levels=["l1", "l2"],
    recurrent_units=10,
    dense_units=5,
)


def test_set_out_sample_residuals_NotFittedError_when_forecaster_not_fitted():
    """
    Test NotFittedError is raised when forecaster is not fitted.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `set_out_sample_residuals()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_is_not_dict():
    """
    Test TypeError is raised when y_true is not a dict.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = 'not_dict'
    y_pred = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"`y_true` must be a dictionary of numpy ndarrays or pandas Series. "
        f"Got {type(y_true)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_is_not_dict():
    """
    Test TypeError is raised when y_pred is not a dict.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}
    y_pred = 'not_dict'

    err_msg = re.escape(
        f"`y_pred` must be a dictionary of numpy ndarrays or pandas Series. "
        f"Got {type(y_pred)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_pred_and_y_true_keys_do_not_match():
    """
    Test TypeError is raised when y_pred and y_true keys do not match.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l3': np.array([1, 2, 3, 4, 5]), 'l4': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"`y_true` and `y_pred` must have the same keys. "
        f"Got {set(y_true.keys())} and {set(y_pred.keys())}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_contains_no_numpy_ndarrays_or_pandas_series():
    """
    Test TypeError is raised when y_true contains no numpy ndarrays or pandas series.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': 'not_ndarray'}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"Values of `y_true` must be numpy ndarrays or pandas Series. "
        f"Got {type(y_true['l1'])} for series l1."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_contains_no_numpy_ndarrays_or_pandas_series():
    """
    Test TypeError is raised when y_pred contains no numpy ndarrays or pandas series.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': 'not_ndarray'}

    err_msg = re.escape(
        f"Values of `y_pred` must be numpy ndarrays or pandas Series. "
        f"Got {type(y_pred['l1'])} for series l1."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_elements_with_different_lengths():
    """
    Test ValueError is raised when y_true and y_pred have elements with different lengths.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([1, 2])}

    err_msg = re.escape(
        '`y_true` and `y_pred` must have the same length. Got 5 and 2 for series l2.'
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_series_with_different_indexes():
    """
    Test ValueError is raised when y_true and y_pred have series with different indexes.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.is_fitted = True
    y_true = {'l1': pd.Series([1, 2, 3, 4, 5], index=[1, 2, 3, 4, 5])}
    y_pred = {'l1': pd.Series([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "When containing pandas Series, elements in `y_true` and "
        "`y_pred` must have the same index."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_inputs_does_not_match_the_target_levels():
    """
    Test ValueError is raised when inputs does not contain keys that match any level.
    """
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.fit(series=series)
    y_true = {'l3': np.array([1, 2, 3])}
    y_pred = {'l3': np.array([1, 2, 3])}

    err_msg = re.escape(
        f"Provided keys in `y_pred` and `y_true` do not match any of the "
        f"target time series in the forecaster, {forecaster.levels}. Residuals "
        f"cannot be updated."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_no_append():
    """
    Test residuals stored when new residuals length is less than 10_000 and 
    append is False.
    """
    rng = np.random.default_rng(123)
    y_true = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000)), 
        'l2': pd.Series(rng.normal(loc=10, scale=10, size=1000))
    }
    y_pred = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000)), 
        'l2': pd.Series(rng.normal(loc=10, scale=10, size=1000))
    }

    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3, transformer_series=None
    )
    forecaster.fit(series=series)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=False)
    results = {
        'l1': np.sort(forecaster.out_sample_residuals_['l1']),
        'l2': np.sort(forecaster.out_sample_residuals_['l2'])
    }

    expected = {
        'l1': np.sort(y_true['l1'] - y_pred['l1']),
        'l2': np.sort(y_true['l2'] - y_pred['l2']),
    }

    assert forecaster.out_sample_residuals_.keys() == expected.keys()
    for key in results.keys():
        np.testing.assert_array_almost_equal(expected[key], results[key])


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_append():
    """
    Test residuals stored when new residuals length is less than 10_000 and 
    append is True.
    """
    rng = np.random.default_rng(123)
    y_true = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000)),
        'l2': pd.Series(rng.normal(loc=10, scale=10, size=1000))
    }
    y_pred = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=1000)),
        'l2': pd.Series(rng.normal(loc=10, scale=10, size=1000))
    }
    
    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3, transformer_series=None
    )
    forecaster.fit(series=series)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)
    results = {
        'l1': np.sort(forecaster.out_sample_residuals_['l1']),
        'l2': np.sort(forecaster.out_sample_residuals_['l2'])
    }

    residuals_1 = (y_true['l1'] - y_pred['l1'])
    residuals_2 = (y_true['l2'] - y_pred['l2'])
    expected = {
        'l1': np.sort(np.concatenate((residuals_1, residuals_1))),
        'l2': np.sort(np.concatenate((residuals_2, residuals_2)))
    }

    assert forecaster.out_sample_residuals_.keys() == expected.keys()
    for key in results.keys():
        np.testing.assert_array_almost_equal(expected[key], results[key])


def test_set_out_sample_residuals_when_residuals_length_is_greater_than_10000():
    """
    Test length residuals stored when its length is greater than 10_000.
    """
    rng = np.random.default_rng(123)
    y_true = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=50_000)),
        'l2': pd.Series(rng.normal(loc=10, scale=10, size=50_000))
    }
    y_pred = {
        'l1': pd.Series(rng.normal(loc=10, scale=10, size=50_000)),
        'l2': pd.Series(rng.normal(loc=10, scale=10, size=50_000))
    }

    forecaster = ForecasterRnn(
        model_no_exog, levels=["l1", "l2"], lags=3
    )
    forecaster.fit(series=series)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)

    assert list(forecaster.out_sample_residuals_.keys()) == ['l1', 'l2']
    for key in forecaster.out_sample_residuals_.keys():
        assert len(forecaster.out_sample_residuals_[key]) == 10_000


def test_out_sample_residuals_and_in_sample_residuals_equivalence():
    """
    Test out sample residuals are equivalent to in-sample residuals 
    when training data and training predictions are passed.
    """
    forecaster = ForecasterRnn(
        model, levels=["l1", "l2"], lags=3
    )
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    X_train, exog_train, y_train, _ = forecaster.create_train_X_y(series=series, exog=exog)
    y_pred_train = forecaster.regressor.predict(
        x=X_train if exog_train is None else [X_train, exog_train], verbose=0
    )

    y_true = []
    y_pred = []
    for i, step in enumerate(forecaster.steps):
        y_true.append(y_train[:, i, :])
        y_pred.append(y_pred_train[:, i, :])
    
    y_true = np.concatenate(y_true)
    y_pred = np.concatenate(y_pred)
    y_true_dict = {}
    y_pred_dict = {}
    for i, level in enumerate(forecaster.levels):
        y_true_dict[level] = forecaster.transformer_series_[level].inverse_transform(
            y_true[:, i].reshape(-1, 1)
        ).ravel()
        y_pred_dict[level] = forecaster.transformer_series_[level].inverse_transform(
            y_pred[:, i].reshape(-1, 1)
        ).ravel()

    forecaster.set_out_sample_residuals(y_true=y_true_dict, y_pred=y_pred_dict)

    assert forecaster.in_sample_residuals_.keys() == forecaster.out_sample_residuals_.keys()
    for k in forecaster.in_sample_residuals_.keys():
        np.testing.assert_array_almost_equal(
            forecaster.in_sample_residuals_[k], forecaster.out_sample_residuals_[k]
        )



# Unit test _initialize_levels_model_selection_multiseries
# ==============================================================================
import re
import pytest
from sklearn.linear_model import Ridge
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.direct import ForecasterDirectMultiVariate
from skforecast.model_selection._utils import _initialize_levels_model_selection_multiseries
from skforecast.exceptions import IgnoredArgumentWarning

# Fixtures
from ..fixtures_model_selection_multiseries import series_wide_range, series_dict_range


def test_initialize_levels_model_selection_multiseries_TypeError_when_levels_not_list_str_None():
    """
    Test TypeError is raised in _initialize_levels_model_selection_multiseries when 
    `levels` is not a `list`, `str` or `None`.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        regressor = Ridge(random_state=123), lags = 4
    )
    levels = 5

    err_msg = re.escape(
        "`levels` must be a `list` of column names, a `str` of a column "
        "name or `None` when using a forecaster of type "
        "['ForecasterRecursiveMultiSeries', 'ForecasterRnn']. If the forecaster "
        "is of type `ForecasterDirectMultiVariate`, this argument is ignored."
    )
    with pytest.raises(TypeError, match = err_msg):
        _initialize_levels_model_selection_multiseries(
            forecaster = forecaster, 
            series     = series_dict_range,
            levels     = levels
        )


def test_initialize_levels_model_selection_multiseries_IgnoredArgumentWarning_forecaster_multivariate_and_levels():
    """
    Test IgnoredArgumentWarning is raised when levels is not forecaster.level or 
    None in ForecasterDirectMultiVariate.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = Ridge(random_state=123),
                     level     = 'l1',
                     lags      = 2,
                     steps     = 3
                 )
    
    levels = 'not_l1_or_None'
    
    warn_msg = re.escape(
        "`levels` argument have no use when the forecaster is of type "
        "`ForecasterDirectMultiVariate`. The level of this forecaster is "
        "'l1', to predict another level, change the `level` "
        "argument when initializing the forecaster."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        levels = _initialize_levels_model_selection_multiseries(
                     forecaster = forecaster, 
                     series     = series_dict_range,
                     levels     = levels
                 )
        
    assert levels == ['l1']


@pytest.mark.parametrize("series_as_dict",
                         [True, False],
                         ids=lambda series_as_dict: f'series_as_dict: {series_as_dict}')
def test_initialize_levels_model_selection_multiseries_ValueError_when_levels_not_in_series(series_as_dict):
    """
    Test ValueError is raised in _initialize_levels_model_selection_multiseries when 
    `levels` is not present in `series`.
    """

    forecaster = ForecasterRecursiveMultiSeries(
        regressor = Ridge(random_state=123), lags = 4
    )
    levels = ['l3']

    err_msg = re.escape(
        "Levels ['l3'] not found in `series`, available levels are "
        "['l1', 'l2']. Review `levels` argument."
    )
    with pytest.raises(ValueError, match = err_msg):
        _initialize_levels_model_selection_multiseries(
            forecaster = forecaster, 
            series     = series_dict_range if series_as_dict else series_wide_range,
            levels     = levels
        )


@pytest.mark.parametrize("series_as_dict",
                         [True, False],
                         ids=lambda series_as_dict: f'series_as_dict: {series_as_dict}')
@pytest.mark.parametrize("levels, levels_expected",
                         [(None, ['l1', 'l2']), 
                          ('l1', ['l1']),
                          (['l1', 'l2'], ['l1', 'l2'])],
                         ids=lambda lags: f'lags, lags_grid_expected: {lags}')
def test_initialize_levels_model_selection_multiseries_for_all_inputs(series_as_dict, levels, levels_expected):
    """
    Test initialize_levels_model_selection_multiseries when levels is None, 
    str or list.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    
    levels = _initialize_levels_model_selection_multiseries(
                 forecaster = forecaster, 
                 series     = series_dict_range if series_as_dict else series_wide_range,
                 levels     = levels
             )
    
    assert levels == levels_expected



# Unit test get_feature_importances ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import RandomForestRegressor
from lightgbm import LGBMRegressor

from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirectMultiVariate

# Fixtures
from .fixtures_forecaster_direct_multivariate import series as series_2
from .fixtures_forecaster_direct_multivariate import exog as exog_2

series = pd.DataFrame({'l1': pd.Series(np.arange(10, dtype=float)), 
                       'l2': pd.Series(np.arange(10, dtype=float))})
exog = pd.DataFrame({'exog_1': np.arange(10, dtype=float),
                     'exog_2': np.arange(100, 110, dtype=float)})


def test_TypeError_is_raised_when_step_is_not_int():
    """
    Test TypeError is raised when calling get_feature_importances() and step is 
    not an int.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = RandomForestRegressor(random_state=123),
                     level     = 'l1',
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(series=series)

    step = 'not_an_int'

    err_msg = re.escape(f"`step` must be an integer. Got {type(step)}.")
    with pytest.raises(TypeError, match = err_msg):         
        forecaster.get_feature_importances(step=step)


def test_NotFittedError_is_raised_when_forecaster_is_not_fitted():
    """
    Test NotFittedError is raised when calling get_feature_importances() and 
    forecaster is not fitted.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = RandomForestRegressor(random_state=123),
                     level     = 'l1',
                     lags      = 3,
                     steps     = 1
                 )

    err_msg = re.escape(
                ("This forecaster is not fitted yet. Call `fit` with appropriate "
                 "arguments before using `get_feature_importances()`.")
              )
    with pytest.raises(NotFittedError, match = err_msg):         
        forecaster.get_feature_importances(step=1)


@pytest.mark.parametrize("step", [0, 2], ids=lambda step: f'step: {step}')
def test_ValueError_is_raised_when_step_is_greater_than_forecaster_max_step(step):
    """
    Test ValueError is raised when calling get_feature_importances() and step is 
    less than 1 or greater than the forecaster.max_step.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = RandomForestRegressor(random_state=123),
                     level     = 'l1',
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(series=series)

    err_msg = re.escape(
                (f"The step must have a value from 1 to the maximum number of steps "
                 f"({forecaster.max_step}). Got {step}.")
            )
    with pytest.raises(ValueError, match = err_msg):         
        forecaster.get_feature_importances(step=step)


def test_output_get_feature_importances_when_regressor_is_RandomForestRegressor_lags_3_step_1():
    """
    Test output of get_feature_importances for step 1, when regressor is 
    RandomForestRegressor with lags=3.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = RandomForestRegressor(random_state=123),
                     level     = 'l1',
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(series=series)

    results = forecaster.get_feature_importances(step=1)
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3'],
                   'importance': np.array([0.17522374, 0.17408016, 0.17918514, 
                                           0.13738555, 0.17994996, 0.15417544])
               }).sort_values(by='importance', ascending=False)
    
    pd.testing.assert_frame_equal(results, expected)
  
    
def test_output_get_feature_importances_when_regressor_is_RandomForestRegressor_lags_3_step_2_exog_included():
    """
    Test output of get_feature_importances for step 2, when regressor is 
    RandomForestRegressor with lags=3 and exog.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = RandomForestRegressor(random_state=123),
                     level     = 'l2',
                     lags      = 3,
                     steps     = 2
                 )
    forecaster.fit(series=series, exog=exog)

    results = forecaster.get_feature_importances(step=2)
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 
                               'exog_1', 'exog_2'],
                   'importance': np.array([0.08793949, 0.11408644, 0.16284783, 
                                           0.15234993, 0.12713124, 0.08502362, 
                                           0.16050297, 0.11011847])
               }).sort_values(by='importance', ascending=False)
    
    pd.testing.assert_frame_equal(results, expected)
    
    
def test_output_get_feature_importances_when_regressor_is_LinearRegression_lags_3_step_1():
    """
    Test output of get_feature_importances for step 1, when regressor is 
    LinearRegression with lags=3.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = LinearRegression(),
                     level     = 'l2',
                     lags      = 3,
                     steps     = 1
                 )
    forecaster.fit(series=series)

    results = forecaster.get_feature_importances(step=1, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3'],
                   'importance': np.array([0.16666667, 0.16666667, 0.16666667, 
                                           0.16666667, 0.16666667, 0.16666667])
               })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression_lags_3_step_2_exog_included():
    """
    Test output of get_feature_importances for step 2, when regressor is 
    LinearRegression with lags=3 and exog.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = LinearRegression(),
                     level     = 'l1',
                     lags      = 3,
                     steps     = 2
                 )
    forecaster.fit(series=series, exog=exog)

    results = forecaster.get_feature_importances(step=2, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 
                               'exog_1', 'exog_2'],
                   'importance': np.array([0.04444444, 0.04444444, 0.04444444,
                                           0.04444444, 0.04444444, 0.04444444, 
                                           0.12765695, 0.12765695])
               })
    
    pd.testing.assert_frame_equal(results, expected)
    

def test_output_get_feature_importances_when_regressor_no_attributes():
    """
    Test output of get_feature_importances when regressor is MLPRegressor with lags=5
    and it is trained with y=pd.Series(np.arange(10)). Since MLPRegressor hasn't attributes
    `feature_importances_` or `coef_, results = None and a warning is raised`
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = MLPRegressor(solver='lbfgs', max_iter=50, random_state=123),
                     level     = 'l1',
                     lags      = 5,
                     steps     = 1
                 )
    forecaster.fit(series=series)

    estimator = forecaster.regressor
    expected = None

    warn_msg = re.escape(
            (f"Impossible to access feature importances for regressor of type "
             f"{type(estimator)}. This method is only valid when the "
             f"regressor stores internally the feature importances in the "
             f"attribute `feature_importances_` or `coef_`.")
        )
    with pytest.warns(UserWarning, match = warn_msg):
        results = forecaster.get_feature_importances(step=1)
        assert results is expected


def test_output_get_feature_importances_when_pipeline_LinearRegression():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + LinearRegression with lags=3).
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor          = make_pipeline(StandardScaler(), 
                                                        LinearRegression()),
                     level              = 'l1',
                     lags               = 3,
                     steps              = 1,
                     transformer_series = None
                 )
    forecaster.fit(series=series)
    results = forecaster.get_feature_importances(step=1, sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3'],
                   'importance': np.array([0.3333333333333332, 0.33333333333333337, 0.33333333333333326, 
                                           0.33333333333333326, 0.33333333333333326, 0.33333333333333326])
               })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_pipeline_RandomForestRegressor():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + RandomForestRegressor with lags=3).
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor          = make_pipeline(StandardScaler(), 
                                                        RandomForestRegressor(random_state=123)),
                     level              = 'l1',
                     lags               = 3,
                     steps              = 1,
                     transformer_series = None
                 )
    forecaster.fit(series=series)
    results = forecaster.get_feature_importances(step=1)
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3'],
                   'importance': np.array([0.17522374, 0.17408016, 0.17918514, 
                                           0.13738555, 0.17994996, 0.15417544])
               }).sort_values(by='importance', ascending=False)
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression_lags_dict_step_2_exog_included():
    """
    Test output of get_feature_importances for step 2, when regressor is 
    LinearRegression with lags as dict and exog.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = LinearRegression(),
                     level     = 'l1',
                     lags      = {'l1': None, 'l2': 3},
                     steps     = 2
                 )
    forecaster.fit(series=series, exog=exog)

    results = forecaster.get_feature_importances(step=2, sort_importance=False)
    expected = pd.DataFrame({
        'feature': ['l2_lag_1', 'l2_lag_2', 'l2_lag_3', 
                    'exog_1', 'exog_2'],
        'importance': np.array([0.05128205128205132, 0.05128205128205135, 0.05128205128205134,
                                0.14729647811635985, 0.14729647811635985])
    })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression_lags_dict_None_step_2_exog_included():
    """
    Test output of get_feature_importances for step 2, when regressor is 
    LinearRegression with lags as dict with None and exog.
    """
    forecaster = ForecasterDirectMultiVariate(
                     regressor = LinearRegression(),
                     level     = 'l1',
                     lags      = {'l1': 3, 'l2': None},
                     steps     = 2
                 )
    forecaster.fit(series=series, exog=exog)

    results = forecaster.get_feature_importances(step=2, sort_importance=False)
    expected = pd.DataFrame({
        'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                    'exog_1', 'exog_2'],
        'importance': np.array([0.05128205128205132, 0.05128205128205135, 0.05128205128205134,
                                0.14729647811635985, 0.14729647811635985])
    })
    
    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_window_features():
    """
    Test output of get_feature_importances when regressor is LGMBRegressor with 
    lags=3 and window features.
    """

    rolling = RollingFeatures(stats=['mean', 'sum'], window_sizes=[3, 5])
    forecaster = ForecasterDirectMultiVariate(
                     regressor       = LGBMRegressor(verbose=-1, random_state=123),
                     level           = 'l1',
                     steps           = 2,
                     lags            = 3,
                     window_features = rolling
                 )
    forecaster.fit(series=series_2, exog=exog_2['exog_1'])

    results = forecaster.get_feature_importances(step=2, sort_importance=False)
    results = results.astype({'importance': float})
    expected = pd.DataFrame({
                   'feature': ['l1_lag_1', 'l1_lag_2', 'l1_lag_3', 
                               'l1_roll_mean_3', 'l1_roll_sum_5',
                               'l2_lag_1', 'l2_lag_2', 'l2_lag_3',
                               'l2_roll_mean_3', 'l2_roll_sum_5',
                               'exog_1'],
                   'importance': np.array([0, 18, 11,  0, 19, 12,  9,  8,  7,  3, 13])
               }).astype({'importance': float})
    
    pd.testing.assert_frame_equal(results, expected)
